"""Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import roc_auc_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC)
from prediction scores.

The return values represent how well the model used is predicting the correct classes, based on the input data. A score of `0.5` means that the model is predicting exactly at chance, i.e. the model's predictions are correct at the same rate as if the predictions were being decided by the flip of a fair coin or the roll of a fair die. A score above `0.5` indicates that the model is doing better than chance, while a score below `0.5` indicates that the model is doing worse than chance.


Note: this implementation can be used with binary, multiclass and
multilabel classification, but some restrictions apply (see Args):
    - binary: The case in which there are only two different label classes, and each example gets only one label. This is the default implementation.
    - multiclass: The case in which there can be more than two different label classes, but each example still gets only one label.
    - multilabel: The case in which there can be more than two different label classes, and each example can have more than one label.

"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Target prediction scores.

        * In the binary case, it corresponds to an array of shape
          `(n_samples,)`. Both probability estimates and non-thresholded
          decision values can be provided. The probability estimates correspond
          to the **probability of the class with the greater label**,
          i.e. `estimator.classes_[1]` and thus
          `estimator.predict_proba(X, y)[:, 1]`. The decision values
          corresponds to the output of `estimator.decision_function(X, y)`.
          See more information in the :ref:`User guide <roc_auc_binary>`;
        * In the multiclass case, it corresponds to an array of shape
          `(n_samples, n_classes)` of probability estimates provided by the
          `predict_proba` method. The probability estimates **must**
          sum to 1 across the possible classes. In addition, the order of the
          class scores must correspond to the order of ``labels``,
          if provided, or else to the numerical or lexicographical order of
          the labels in ``y_true``. See more information in the
          :ref:`User guide <roc_auc_multiclass>`;
        * In the multilabel case, it corresponds to an array of shape
          `(n_samples, n_classes)`. Probability estimates are provided by the
          `predict_proba` method and the non-thresholded decision values by
          the `decision_function` method. The probability estimates correspond
          to the **probability of the class with the greater label for each
          output** of the classifier. See more information in the
          :ref:`User guide <roc_auc_multilabel>`.

    references: True labels or binary label indicators. The binary and
        multiclass cases expect labels with shape (n_samples,) while the
        multilabel case expects binary label indicators with shape
        (n_samples, n_classes).

    average (optional): {'micro', 'macro', 'samples','weighted'} or None,
        default='macro'.
        If ``None``, the scores for each class are returned. Otherwise,
        this determines the type of averaging performed on the data:
        Note: multiclass ROC AUC currently only handles the 'macro' and
        'weighted' averages.

        ``'micro'``:
            Calculate metrics globally by considering each molflux of the label
            indicator matrix as a label.

        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.

        ``'weighted'``:
            Calculate metrics for each label, and find their average, weighted
            by support (the number of true instances for each label).

        ``'samples'``:
            Calculate metrics for each instance, and find their average.

        Will be ignored when ``y_true`` is binary.

    sample_weight (optional): Weighting of each sample.

    max_fpr (optional): If not ``None``, the standardized partial AUC [2]_ over
        the range [0, max_fpr] is returned. For the multiclass case,
        ``max_fpr``, should be either equal to ``None`` or ``1.0`` as AUC ROC
        partial computation currently is not supported for multiclass.

    multi_class (optional): {'raise', 'ovr', 'ovo'}, default='raise'.
        Only used for multiclass targets. Determines the type of configuration
        to use. The default value raises an error, so either
        ``'ovr'`` or ``'ovo'`` must be passed explicitly.

        ``'ovr'``:
            Stands for One-vs-rest. Computes the AUC of each class
            against the rest [3]_ [4]_. This
            treats the multiclass case in the same way as the multilabel case.
            Sensitive to class imbalance even when ``average == 'macro'``,
            because class imbalance affects the composition of each of the
            'rest' groupings.
        ``'ovo'``:
            Stands for One-vs-one. Computes the average AUC of all
            possible pairwise combinations of classes [5]_.
            Insensitive to class imbalance when
            ``average == 'macro'``.

    labels (optional): Only used for multiclass targets. List of labels that
        index the classes in ``predictions``. If ``None``, the numerical or
        lexicographical order of the labels in ``references`` is used. Defaults
        to ``None``.

Returns:
    roc_auc: Area Under the Receiver Operating Characteristic Curve.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("roc_auc")
    >>> references = [1, 0, 1, 1, 0, 0]
    >>> predictions = [0.5, 0.2, 0.99, 0.3, 0.1, 0.7]
    >>> results = metric.compute(references=references, predictions=predictions)
    >>> print(round(results['roc_auc'], 2))
    0.78
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("roc_auc", config_name="multiclass")
    >>> references = [1, 0, 1, 2, 2, 0]
    >>> predictions = [[0.3, 0.5, 0.2],
    ...                 [0.7, 0.2, 0.1],
    ...                 [0.005, 0.99, 0.005],
    ...                 [0.2, 0.3, 0.5],
    ...                 [0.1, 0.1, 0.8],
    ...                 [0.1, 0.7, 0.2]]
    >>> results = metric.compute(references=references, predictions=predictions, multi_class='ovr')
    >>> print(round(results['roc_auc'], 2))
    0.85
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("roc_auc", config_name="multilabel")
    >>> references = [[1, 1, 0],
    ...         [1, 1, 0],
    ...         [0, 1, 0],
    ...         [0, 0, 1],
    ...         [0, 1, 1],
    ...         [1, 0, 1]]
    >>> predictions = [[0.3, 0.5, 0.2],
    ...                 [0.7, 0.2, 0.1],
    ...                 [0.005, 0.99, 0.005],
    ...                 [0.2, 0.3, 0.5],
    ...                 [0.1, 0.1, 0.8],
    ...                 [0.1, 0.7, 0.2]]
    >>> results = metric.compute(references=references, predictions=predictions, average=None)
    >>> print([round(res, 2) for res in results['roc_auc']])
    [0.83, 0.38, 0.94]
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class RocAuc(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Value("int32"),
                }
                if self.config_name == "multiclass"
                else {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        average: str = "macro",
        max_fpr: float | None = None,
        multi_class: str = "raise",
        labels: list[str] | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        score = roc_auc_score(
            y_true=references,
            y_score=predictions,
            average=average,
            sample_weight=sample_weight,
            max_fpr=max_fpr,
            multi_class=multi_class,
            labels=labels,
        )
        return {self.tag: score}



import molflux.features
from molflux.datasets.interfaces import Representations


def test_integrates_with_molflux_features():
    """That the Representations interface used by datasets is compatible
    with the Representations defined by molflux.features.

    That is, we check that datasets works with Representations from molflux.features.
    """
    assert isinstance(molflux.features.Representations, Representations)



from contextlib import nullcontext
from typing import ContextManager

import pandas as pd
import pytest
import torch

import datasets
from molflux.modelzoo.load import load_model
from molflux.modelzoo.protocols import Model

model_name = "lightning_mlp_regressor"

_X_FEATURES = ["X_col_1", "X_col_2"]
_Y_FEATURES = ["y_col"]


@pytest.fixture()
def train_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            [1.0, 2.0, 3.0],
            [1.0, 3.0, 4.0],
            [2.0, 4.0, 6.0],
            [2.0, 4.0, 6.0],
            [2.0, 4.0, 6.0],
        ],
        columns=_X_FEATURES + _Y_FEATURES,
    )


@pytest.fixture()
def validation_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            [1.0, 2.0, 3.0],
            [1.0, 3.0, 4.0],
            [2.0, 4.0, 6.0],
            [2.0, 4.0, 6.0],
            [2.0, 5.0, 7.0],
        ],
        columns=_X_FEATURES + _Y_FEATURES,
    )


@pytest.fixture()
def predict_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            [1.0, 2.0],
            [1.0, 3.0],
            [2.0, 4.0],
            [2.0, 4.0],
            [2.0, 5.0],
        ],
        columns=_X_FEATURES,
    )


@pytest.fixture()
def empty_predict_df() -> pd.DataFrame:
    return pd.DataFrame([], columns=_X_FEATURES)


@pytest.fixture()
def train_dataset(train_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(train_df)


@pytest.fixture()
def train_dataset_b(train_df: pd.DataFrame) -> datasets.Dataset:
    joint_df = pd.concat([train_df, train_df]).iloc[:8]
    return datasets.Dataset.from_pandas(joint_df)


@pytest.fixture()
def validation_dataset(validation_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(validation_df)


@pytest.fixture()
def predict_dataset(predict_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(predict_df)


@pytest.fixture()
def empty_predict_dataset(empty_predict_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(empty_predict_df)


@pytest.fixture(scope="function")
def fixture_model() -> Model:
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        hidden_dim=3,
        input_dim=2,
        num_tasks=len(_Y_FEATURES),
    )


@pytest.fixture(scope="function")
def fixture_multi_layer_model() -> Model:
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        hidden_dim=3,
        num_layers=4,
        input_dim=2,
        num_tasks=len(_Y_FEATURES),
    )


@pytest.fixture(scope="function")
def fixture_compiled_model() -> Model:
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        hidden_dim=3,
        input_dim=2,
        num_tasks=len(_Y_FEATURES),
    )


@pytest.fixture(scope="function")
def fixture_pre_trained_model() -> Model:
    model = load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        hidden_dim=3,
        input_dim=2,
        num_tasks=len(_Y_FEATURES),
    )
    # set params to nice numbers
    model.module = model._instantiate_module()  # type: ignore
    for n, p in model.module.named_parameters():  # type: ignore
        if n == "module.0.weight":
            p.data = torch.ones_like(p) * 0
        elif n == "module.0.bias":
            p.data = torch.ones_like(p) * 1
        elif n == "module.2.weight":
            p.data = torch.ones_like(p) * 2
        elif n == "module.2.bias":
            p.data = torch.ones_like(p) * 3
        else:
            raise KeyError("Param not found in module")

    return model


@pytest.fixture(scope="function")
def fixture_pre_trained_multi_layer_model() -> Model:
    model = load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        hidden_dim=3,
        num_layers=4,
        input_dim=2,
        num_tasks=len(_Y_FEATURES),
    )
    # set params to nice numbers
    model.module = model._instantiate_module()  # type: ignore
    for n, p in model.module.named_parameters():  # type: ignore
        if n == "module.2.weight":
            p.data = torch.ones_like(p) * 0
        elif n == "module.2.bias":
            p.data = torch.ones_like(p) * 1

    return model


@pytest.fixture()
def compile_error_context(compile: bool) -> ContextManager:
    if not compile:
        return nullcontext()
    else:
        return pytest.raises(ValueError, match=r"^Compilation temporarily disabled")

    # TODO enable after PyTorch 2.1 is available
    # if sys.version_info >= (3, 11):
    #     return pytest.raises(
    #         RuntimeError,
    #         match=r"^Python 3.11\+ not yet supported for torch.compile$",
    #     )
    # elif sys.version_info.major == 3 and sys.version_info.minor == 8:
    #     return pytest.raises(
    #         (AssertionError, torch._dynamo.exc.BackendCompilerFailed),
    #     )
    # else:
    #     return nullcontext()



from typing import Any

try:
    from rdkit.Chem.AtomPairs.Pairs import GetAtomPairFingerprint

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The atom-pair fingerprint for a molecule.

The atom-pair (and related topological-torsion) fingerprints, are implemented
based on the original paper. Atoms are typed based on atomic number,
number of pi electrons, and the degree of the atom. Optionally information
about atomic chirality can also be integrated into the atom types. Both
fingerprint types can be generated in explicit or sparse form and as bit or
count vectors. These fingerprint types are different from the others in the
RDKit in that bits in the sparse form of the fingerprint can be directly
explained (i.e. the “hashing function” used is fully reversible).

These fingerprints were originally “intended” to be used in count-vectors and
they seem to work better that way. The default behavior of the explicit
bit-vector forms of both fingerprints is to use a “count simulation” procedure
where multiple bits are set for a given feature if it occurs more than once.
The default behavior is to use 4 fingerprint bits for each feature
(so a 2048 bit fingerprint actually stores information about the same number of
features as a 512 bit fingerprint that isn`t using count simulation). The bins
correspond to counts of 1, 2, 4, and 8. As an example of how this works: if a
feature occurs 5 times in a molecule, the bits corresponding to counts 1, 2,
and 4 will be set.

The algorithm used is described here: R.E. Carhart, D.H. Smith, R. Venkataraghavan; "Atom Pairs as Molecular Features in Structure-Activity Studies: Definition and Applications" JCICS 25, 64-73 (1985).
"""


class AtomPairUnfolded(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        min_length: int = 1,
        max_length: int = 30,
        from_atoms: list[int] | None = None,
        ignore_atoms: list[int] | None = None,
        atom_invariants: list[int] | None = None,
        include_chirality: bool = False,
        use_2d: bool = True,
        conf_id: int = -1,
        **kwargs: Any,
    ) -> dict[str, list[dict[str, int]]]:
        """Generates an atom-pair fingerprint for each input molecule.

        Args:
            samples: The molecules to be fingerprinted.
            min_length: The minimum distance between atoms to be considered in
                a pair. Default is 1 bond.
            max_length: The maximum distance between atoms to be considered in
                a pair. Default is 30 bonds.
            from_atoms: If not None, only atom pairs that involve the specified
                atoms will be included in the fingerprint. Defaults to `None`.
            ignore_atoms: If not None, any atom pairs that include the specified
                atoms will not be included in the fingerprint. Defaults to
                `None`.
            atom_invariants: A list of invariants to use for the atom hashes.
                Note that only the first 9 bits of each invariant are used.
                Defaults to `None`.
            include_chirality: If `True`, chirality will be used in the atom
                invariants. Defaults to `False`. Note that this is ignored if
                `atom_invariants` are provided.
            use_2d: Whether to use the 2D (topological) distance matrix.
                Defaults to `True`.
            conf_id: The conformation to use if 3D distances are being used.
                Defaults to `-1`.

        Returns:
            Unfolded Atom-pair fingerprints, as dictionaries.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('atom_pair_unfolded')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples, max_length=16)
            {'atom_pair_unfolded': [{'689473': 6, '689474': 6, '689475': 3}]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        atom_pairs_fp_list: list[dict[str, int]] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = GetAtomPairFingerprint(
                    mol,
                    minLength=min_length,
                    maxLength=max_length,
                    fromAtoms=from_atoms or [],
                    ignoreAtoms=ignore_atoms or [],
                    atomInvariants=atom_invariants or [],
                    includeChirality=include_chirality,
                    use2D=use_2d,
                    confId=conf_id,
                )
                atom_pairs_fp_list.append(
                    {str(k): v for k, v in rd_fp.GetNonzeromolfluxs().items()},
                )

        return {self.tag: atom_pairs_fp_list}



import numpy as np
import pandas as pd
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "ordered_split"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_inputs():
    n = 100
    dataset = np.random.rand(n)
    groups = np.concatenate([np.arange(n - 20), np.array([None] * 20)])
    np.random.shuffle(groups)
    return dataset, groups


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_requires_groups(fixture_test_strategy):
    """That you need to provide groups to use this strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset))


def test_hard_coded_example(fixture_test_strategy):
    """Test to illustrate the split"""
    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [2, 7, 5, 6, 3, 1, 0, 8, 9, 4]

    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)
    train_indices, validation_indices, test_indices = next(indices)

    assert train_indices == [6, 5, 0, 4, 9, 2, 3, 1]
    assert validation_indices == [7]
    assert test_indices == [8]


def test_hard_coded_example_with_nones(fixture_test_strategy):
    """Test to illustrate the split when there are nones:
    Nones will be added to train and ignored when computing fractions
    """

    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [3, 0, 1, None, 4, 5, 2, 6, None, 9]

    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, same_group_same_set=False)
    train_indices, validation_indices, test_indices = next(indices)

    assert train_indices == [3, 8, 1, 2, 6, 0, 4, 5]
    assert validation_indices == [7]
    assert test_indices == [9]


def test_example_with_nulls_nones(fixture_test_strategy):
    """Test to illustrate the split when there are nones or other null values:
    Nones will be added to train and ignored when computing fractions
    """

    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [3, 0, 1, np.nan, 4, 5, 2, 6, pd.NaT, 9]

    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, same_group_same_set=False)
    train_indices, validation_indices, test_indices = next(indices)

    assert train_indices == [3, 8, 1, 2, 6, 0, 4, 5]
    assert validation_indices == [7]
    assert test_indices == [9]


def test_hard_coded_example_with_non_unique_groups(fixture_test_strategy):
    """Test to illustrate the split when there are nones:
    Nones will be added to train and ignored when computing fractions
    """

    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [0, 0, 0, 0, 1, 1, 1, 2, 2, 3]

    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, same_group_same_set=False)
    train_indices, validation_indices, test_indices = next(indices)

    assert train_indices == [0, 1, 2, 3, 4, 5, 6, 7]
    assert validation_indices == [8]
    assert test_indices == [9]


def test_hard_coded_example_with_unusual_ratios(fixture_test_strategy):
    """Test to illustrate the split when there are nones:
    Nones will be added to train and ignored when computing fractions
    """

    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [2, 0, 1, None, 2, 5, 2, 5, None, 9]

    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        groups=groups,
        train_fraction=0.1,
        validation_fraction=0.8,
        test_fraction=0.1,
        undefined_groups_in_train=False,
        same_group_same_set=False,
    )
    train_indices, validation_indices, test_indices = next(indices)

    assert train_indices == [3]
    assert validation_indices == [8, 1, 2, 0, 4, 6, 5, 7]
    assert test_indices == [9]


def test_example_same_test_ratios_but_different_train_val(
    fixture_sample_inputs,
    fixture_test_strategy,
):
    """Test to illustrate that the same test ratio with different train val results
    in the same test set size (consistency)
    """

    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy

    indices_1 = strategy.split(
        dataset=dataset,
        groups=groups,
        train_fraction=0.6,
        validation_fraction=0.2,
        test_fraction=0.2,
        same_group_same_set=False,
    )
    _, _, test_indices_1 = next(indices_1)

    indices_2 = strategy.split(
        dataset=dataset,
        groups=groups,
        train_fraction=0.3,
        validation_fraction=0.5,
        test_fraction=0.2,
        same_group_same_set=False,
    )
    _, _, test_indices_2 = next(indices_2)

    assert len(test_indices_1) == len(test_indices_2)


def test_n_splits(fixture_test_strategy):
    """Test the n_splits behaviour. should return n_splits folds with the required fractions"""

    n = 100
    dataset = np.random.rand(n)
    groups = np.arange(n)
    np.random.shuffle(groups)

    train_fraction = [0.6, 0.7, 0.8]
    gap_train_validation_fraction = [0.05, 0.05, 0.05]
    validation_fraction = [0.2, 0.1, 0.0]
    gap_validation_test_fraction = [0.05, 0.05, 0.05]
    test_fraction = [0.1, 0.1, 0.1]

    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        groups=groups,
        train_fraction=train_fraction,
        gap_train_validation_fraction=gap_train_validation_fraction,
        validation_fraction=validation_fraction,
        gap_validation_test_fraction=gap_validation_test_fraction,
        test_fraction=test_fraction,
        same_group_same_set=False,
    )

    for ii, (train_indices, validation_indices, test_indices) in enumerate(indices):
        assert len(train_indices) == round(train_fraction[ii] * len(dataset))
        assert len(validation_indices) == round(validation_fraction[ii] * len(dataset))
        assert len(test_indices) == round(test_fraction[ii] * len(dataset))


def test_splits_are_disjoint(fixture_sample_inputs, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        groups=groups,
        same_group_same_set=False,
    )
    for train_indices, validation_indices, test_indices in indices:
        assert (
            set(train_indices).isdisjoint(set(validation_indices))
            and set(validation_indices).isdisjoint(set(test_indices))
            and set(train_indices).isdisjoint(set(test_indices))
        )


def test_deterministic(fixture_sample_inputs, fixture_test_strategy):
    """That repeated splits are deterministic."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of ordered splits
    indices_a = strategy.split(dataset, groups=groups, same_group_same_set=False)
    train_indices_a1, validation_indices_a1, test_indices_a1 = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, groups=groups, same_group_same_set=False)
    train_indices_b1, validation_indices_b1, test_indices_b1 = next(indices_b)

    # Check that we get the same results across folds
    assert train_indices_a1 == train_indices_b1
    assert validation_indices_a1 == validation_indices_b1
    assert test_indices_a1 == test_indices_b1


def test_ignoring_nones_ratios(fixture_sample_inputs, fixture_test_strategy):
    """That ignoring none for computing split ratios means more data in train and less in val and test"""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of ordered splits
    indices_a = strategy.split(
        dataset,
        groups=groups,
        undefined_groups_in_train=True,
        same_group_same_set=False,
    )
    train_indices_a1, validation_indices_a1, test_indices_a1 = next(indices_a)

    indices_b = strategy.split(
        dataset,
        groups=groups,
        undefined_groups_in_train=False,
        same_group_same_set=False,
    )
    train_indices_b1, validation_indices_b1, test_indices_b1 = next(indices_b)

    # assert that there are more in train and less in val and test
    assert len(train_indices_b1) < len(train_indices_a1)
    assert (len(validation_indices_b1) > len(validation_indices_a1)) or (
        len(test_indices_b1) > len(test_indices_a1)
    )


def test_same_groups_same_set(fixture_test_strategy):
    """Test to check that same_groups are put in the same set (no overlap)"""
    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [0, 1, 1, 1, 2, 2, 2, 2, 2, 3]

    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)
    train_indices, validation_indices, test_indices = next(indices)

    assert train_indices == [0, 1, 2, 3, 4, 5, 6, 7, 8]
    assert validation_indices == []
    assert test_indices == [9]


def test_min_size_test(fixture_test_strategy):
    """Test to check that min_test_size is respected"""
    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    groups = [0, 1, 1, 1, 2, 2, 2, 2, 2, 3]

    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, min_test_size=5)
    train_indices, validation_indices, test_indices = next(indices)

    assert len(test_indices) > 5


def test_n_splits_with_min_test_size(fixture_test_strategy):
    """Test the n_splits behaviour with min_test_size list"""

    n = 100
    dataset = np.random.rand(n)
    groups = np.arange(n)
    np.random.shuffle(groups)

    train_fraction = [0.6, 0.7, 0.8]
    gap_train_validation_fraction = [0.05, 0.05, 0.05]
    validation_fraction = [0.2, 0.1, 0.0]
    gap_validation_test_fraction = [0.05, 0.05, 0.05]
    test_fraction = [0.1, 0.1, 0.1]
    min_test_size = [0, 5, 9]

    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        groups=groups,
        train_fraction=train_fraction,
        gap_train_validation_fraction=gap_train_validation_fraction,
        validation_fraction=validation_fraction,
        gap_validation_test_fraction=gap_validation_test_fraction,
        test_fraction=test_fraction,
        same_group_same_set=False,
        min_test_size=min_test_size,
    )

    for ii, (_train_indices, _validation_indices, test_indices) in enumerate(indices):
        assert len(test_indices) >= min_test_size[ii]



import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.accuracy import Accuracy


@pytest.fixture()
def fixture_metric():
    return load_metric("accuracy")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "accuracy" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, Accuracy)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_integer_labels(fixture_metric):
    """That can use integer labels."""
    metric = fixture_metric
    predictions = [0, 2, 1, 3]
    references = [0, 1, 2, 3]
    result = metric.compute(predictions=predictions, references=references)
    assert result["accuracy"] == 0.5


def test_normalised_accuracy(fixture_metric):
    """That can calculate normalised accuracy."""
    metric = fixture_metric
    predictions = [0, 2, 1, 3]
    references = [0, 1, 2, 3]
    result = metric.compute(
        predictions=predictions,
        references=references,
        normalize=False,
    )
    assert result["accuracy"] == 2


def test_perfect_accuracy(fixture_metric):
    """That get unitary accuracy for perfect predictions."""
    metric = fixture_metric
    predictions = [0, 1]
    references = [0, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["accuracy"] == 1.0


def test_binary_multilabel():
    """Multilabel case with binary label indicators."""
    metric = load_metric("accuracy", config_name="multilabel")
    predictions = np.array([[0, 1], [1, 1]])
    references = np.ones((2, 2))
    result = metric.compute(predictions=predictions, references=references)
    assert result["accuracy"] == 0.5



from colorama import Fore, Style


class ExtrasDependencyImportError(Exception):
    """Raisable on ImportErrors due to missing package extras."""

    def __init__(self, extras_type: str, nested_error: Exception):
        message = (
            f"\n\n📦 {nested_error}\n\n"
            + f"{Style.DIM}# have you tried running the following?{Style.RESET_ALL}\n"
            + f"$ {Style.BRIGHT + Fore.GREEN}pip install 'molflux[{extras_type}]'{Style.RESET_ALL}"
        )

        super().__init__(message)


class OptionalDependencyImportError(Exception):
    """Raisable on ImportErrors due to missing optional dependencies."""

    def __init__(self, dependency: str, package: str):
        message = (
            f"Optional dependency {dependency} missing."
            + f"{Style.DIM}# have you tried running the following?{Style.RESET_ALL}\n"
            + f"$ {Style.BRIGHT + Fore.GREEN}pip install '{package}'{Style.RESET_ALL}"
        )

        super().__init__(message)


class NotTrainedError(ValueError, AttributeError):
    """Exception class to raise if a model is used before fitting.

    This class inherits from both ValueError and AttributeError to help with
    exception handling.

    References:
        https://scikit-learn.org/stable/modules/generated/sklearn.exceptions.NotFittedError.html
    """

    def __init__(self, message: str | None = None) -> None:
        default_message = "This estimator has not been trained yet: please train the estimator with appropriate arguments."
        super().__init__(message or default_message)



import itertools
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

import pytest

import datasets
import molflux.core.tracking
import molflux.modelzoo
import molflux.splits

parametrized_output_directory_fixtures = [
    "tmp_path",
    "fixture_test_bucket",
    "fixture_nested_tmp_path",
    "fixture_nested_test_bucket",
]

file_formats = ["parquet", "csv", "json"]


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_log_params_creates_json(output_directory_fixture, request):
    """That can log arbitrary param payloads to json."""
    output_directory = request.getfixturevalue(output_directory_fixture)

    params = {"a": [1, 2, 3], "b": "value"}
    output_path: Path = output_directory / "params.json"

    molflux.core.tracking.log_params(params, path=output_path)
    assert output_path.is_file()


# TODO(avianello) mock the datasets S3 callstack to test saving datasets on s3 too
@pytest.mark.parametrize(
    "output_directory_fixture, format",
    itertools.product(parametrized_output_directory_fixtures[:1], file_formats),
)
def test_log_dataset_saves_dataset_as_persisted_file_format(
    output_directory_fixture,
    format,
    request,
):
    """That can log an arbitrary dataset as persisted file in a variety of formats."""
    output_directory = request.getfixturevalue(output_directory_fixture)

    dataset = datasets.Dataset.from_dict({"x1": [1, 2, 3], "x2": [1, 2, 3]})
    output_path: Path = output_directory / f"data.{format}"

    molflux.core.tracking.log_dataset(dataset, path=output_path)
    assert output_path.is_file()


# TODO(avianello) mock the datasets S3 callstack to test saving datasets on s3 too
@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures[:1],
)
def test_log_featurised_dataset_saves_dataset_as_canonical_file(
    output_directory_fixture,
    request,
):
    """That featurised datasets are logged according to standard format."""
    output_directory = request.getfixturevalue(output_directory_fixture)

    dataset = datasets.Dataset.from_dict({"x1": [1, 2, 3], "x2": [1, 2, 3]})
    molflux.core.tracking.log_featurised_dataset(dataset, directory=output_directory)

    expected_output_path = output_directory / "featurised_dataset.parquet"
    assert expected_output_path.is_file()


# TODO(avianello) mock the datasets S3 callstack to test saving datasets on s3 too
@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures[:1],
)
def test_log_fold_saves_fold_as_canonical_files(output_directory_fixture, request):
    """That folds are logged according to standard format."""
    output_directory = request.getfixturevalue(output_directory_fixture)

    split = datasets.Dataset.from_dict({"x1": [1, 2, 3], "x2": [1, 2, 3]})
    fold = datasets.DatasetDict({"train": split, "validation": split, "test": split})
    molflux.core.tracking.log_fold(fold, directory=output_directory)

    expected_output_paths = [
        output_directory / f"{split_name}.parquet" for split_name in fold.keys()
    ]
    for output_path in expected_output_paths:
        assert output_path.is_file()


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_log_model_params_creates_canonical_file(output_directory_fixture, request):
    """That the log_model_params function saves model metadata in a file
    called model_params.json"""
    output_directory = request.getfixturevalue(output_directory_fixture)

    model = molflux.modelzoo.load_model("random_forest_regressor")
    molflux.core.tracking.log_model_params(model, directory=output_directory)

    expected_output_path: Path = output_directory / "model_params.json"
    assert expected_output_path.is_file()


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_log_pipeline_config_creates_canonical_file(output_directory_fixture, request):
    """That the log_pipeline_config function saves pipeline config metadata
    in a file called pipeline.json"""
    output_directory = request.getfixturevalue(output_directory_fixture)

    config = {"a": [1, 2, 3], "b": "value"}
    molflux.core.tracking.log_pipeline_config(config, directory=output_directory)

    expected_output_path: Path = output_directory / "pipeline.json"
    assert expected_output_path.is_file()


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_log_scores_creates_canonical_file(output_directory_fixture, request):
    """That the log_scores function saves model scores in a file called scores.json"""
    output_directory = request.getfixturevalue(output_directory_fixture)

    scores = {
        "train": {"y1": {"r2": 0.99}},
        "validation": {"y1": {"r2": 0.99}},
        "test": {"y1": {"r2": 0.99}},
    }
    molflux.core.tracking.log_scores(scores, directory=output_directory)

    expected_output_path: Path = output_directory / "scores.json"
    assert expected_output_path.is_file()


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_log_splitting_strategy_creates_canonical_file(
    output_directory_fixture,
    request,
):
    """That the log_splitting_strategy function saves splitting strategy
    metadata in a file called splitting_strategy.json"""
    output_directory = request.getfixturevalue(output_directory_fixture)

    splitting_strategy = molflux.splits.load_splitting_strategy("shuffle_split")
    molflux.core.tracking.log_splitting_strategy(
        splitting_strategy,
        directory=output_directory,
    )

    expected_output_path: Path = output_directory / "splitting_strategy.json"
    assert expected_output_path.is_file()



from typing import Any

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import (
    rdkit_mol_from_smiles,
    to_smiles,
)
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

try:
    from rdkit import Chem

except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

_TOXICOPHORE_SMARTS = [
    "O=N(~O)a",
    "a[NH2]",
    "a[N;X2]=O",
    "CO[N;X2]=O",
    "N[N;X2]=O",
    "O1[c,C]-[c,C]1",
    "C1NC1",
    "N=[N+]=[N-]",
    "C=[N+]=[N-]",
    "N=N-N",
    "c[N;X2]!@;=[N;X2]c",
    "[OH,NH2][N,O]",
    "[OH]Na",
    "[Cl,Br,I]C",
    "[Cl,Br,I]C=O",
    "[N,S]!@[C;X4]!@[CH2][Cl,Br,I]",
    "[cH]1[cH]ccc2c1c3c(cc2)cc[cH][cH]3",
    "[cH]1cccc2c1[cH][cH]c3c2ccc[cH]3",
    "[$([C,c]OS(=O)(=O)O!@[c,C]),$([c,C]S(=O)(=O)O!@[c,C])]",
    "O=N(~O)N",
    "[$(O=[CH]C=C),$(O=[CH]C=O)]",
    "[N;v4]#N",
    "O=C1CCO1",
    "[CH]=[CH]O",
    "[NH;!R][NH;R]a",
    "[CH3][NH]a",
    "aN([$([OH]),$(O*=O)])[$([#1]),$(C(=O)[CH3]),$([CH3]),$([OH]),$(O*=O)]",
    "a13~a~a~a~a2~a1~a(~a~a~a~3)~a~a~a~2",
    "a1~a~a~a2~a~1~a~a3~a(~a~2)~a~a~a~3",
    "a1~a~a~a2~a~1~a~a~a3~a~2~a~a~a~3",
    "a1~a~a~a~a2~a~1~a3~a(~a~2)~a~a~a~a~3",
    "a1~a~a~a~a2~a~1~a~a3~a(~a~2)~a~a~a~3",
    "a1~a~a~a~a2~a~1~a~a3~a(~a~2)~a~a~a~a~3",
    "a1~a~a~a~a2~a~1~a~a~a3~a~2~a~a~a~3",
    "a1~a~a~a~a2~a~1~a~a~a3~a~2~a~a~a~a~3",
    "a13~a~a~a~a2~a1~a(~a~a~a~3)~a~a~2",
]

_ALERTS = [Chem.MolFromSmarts(s) for s in _TOXICOPHORE_SMARTS]

_DESCRIPTION = """
Substructure matching against toxicophore SMARTS patterns used in
Pires et al (2015) J. Med. Chem. 2015, 58, 9, 4066-4072
"""


class Toxicophores(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        **kwargs: Any,
    ) -> dict[str, list[list[Any]]]:
        """
        Calculates toxicophore substructure matches.

        Args:
            samples: The molecules for which to calculate descriptors.

        Returns:
            Boolean vector corresponding to absence/presence of substruct
            match for each pattern.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation("toxicophores")
            >>> samples = ["C1NC1"]
            >>> representation.featurise(samples)
            {'toxicophores': [[0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        match_list: list[list] = []

        for sample in samples:
            with featurisation_error_harness(sample):
                smiles = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smiles)
                match = [int(mol.HasSubstructMatch(alert)) for alert in _ALERTS]
                match_list.append(match)

        return {self.tag: match_list}



"""Computes an ROC curve over SMILES strings"""

import logging
from typing import Any

import evaluate
import numpy as np

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_CITATION = """\
@InProceedings{exscientia:validity_roc,
title = {Validity ROC},
authors={Exscientia},
year={2022}
}
"""

_DESCRIPTION = """\
Computes a validity ROC curve over SMILES strings, i.e. "how many SMILES in the first k results are valid"?
"""


# TODO: Add description of the arguments of the metric here
_KWARGS_DESCRIPTION = """
Calculates what proportion of SMILES-encoded predictions are valid within the top k tries.

Args:
    predictions: list of lists of predictions to score. Each prediction should be a SMILES string.
    references:  list of references. Ignored by this metric, but should be of the same length as 'predictions'.
    invalid_string: the invalid string (default "<INVALID>").
    average: If `True`, an molflux-wise average across results is performed.

Returns:
    validity_roc: list whose k component is the answer to "is the correct synthesis product in top-k predictions"
        If 'average=True', "on average, is the correct synthesis product in top-k predictions"

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("validity_roc")
    >>> predictions = [["CC", "CC", None, None, "CCC"]]
    >>> references = [None]
    >>> metric.compute(predictions=predictions, references=references)
    {'validity_roc': [[1.0, 1.0, 0.66..., 0.5, 0.6]]}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class ValidityRoc(HFMetric):
    """Computes the validity ROC curve."""

    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            # This is the description that will appear on the metrics page.
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            # This defines the format of each prediction and reference
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("string")),
                    "references": datasets.Value("string"),
                },
            ),
        )

    def _pre_process_inputs(
        self,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
    ) -> tuple[ArrayLike | None, ...]:
        """Converts None inputs to strings to allow using None as invalid token.

        Otherwise, None molfluxs would raise as they do not count as string features.
        """
        if predictions is not None:
            predictions = [
                [p if p is not None else "None" for p in ps] for ps in predictions
            ]

        if references is not None:
            references = [r if r is not None else "None" for r in references]

        return predictions, references

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        invalid_token: str | None = None,
        average: bool = False,
        **kwargs: Any,
    ) -> MetricResult:
        """Returns the validity ROC curve"""

        if invalid_token is None:
            invalid_token = "None"  # noqa: S105

        # vector of 0./1. representing (in)valid SMILES
        valid_preds = [
            [p != invalid_token for p in one_set_of_preds]
            for one_set_of_preds in predictions
        ]

        def running_mean(x: list[Any]) -> np.ndarray:
            cumsum = np.cumsum(x)
            norm = np.arange(1, np.shape(x)[0] + 1)
            mean: np.ndarray = cumsum / norm
            return mean

        running_means = np.array([running_mean(p) for p in valid_preds])

        if average:
            running_means = running_means.mean(axis=0, keepdims=True)

        validity_roc: list[float] = running_means.tolist()

        return {self.tag: validity_roc}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

try:
    from mhfp.encoder import MHFPEncoder
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

if TYPE_CHECKING:
    from molflux.features.typing import MolArray

_DESCRIPTION = """
MHFP6 (MinHash fingerprint, up to six bonds) is a molecular fingerprint which
encodes detailed substructures using the extended connectivity principle of
ECFP in a fundamentally different manner, increasing the performance of exact
nearest neighbor searches in benchmarking studies and enabling the application
of locality sensitive hashing (LSH) approximate nearest neighbor search
algorithms.

To describe a molecule, MHFP6 extracts the SMILES of all circular substructures
around each atom up to a diameter of six bonds and applies the MinHash method
to the resulting set. MHFP6 outperforms ECFP4 in benchmarking analog recovery
studies. Furthermore, MHFP6 outperforms ECFP4 in approximate nearest neighbor
searches by two orders of magnitude in terms of speed, while decreasing the
error rate

https://jcheminf.biomedcentral.com/articles/10.1186/s13321-018-0321-8
"""


class MHFPUnfolded(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        n_permutations: int = 1024,
        radius: int = 3,
        rings: bool = True,
        kekulise: bool = True,
        sanitise: bool = True,
        seed: int = 42,
        **kwargs: Any,
    ) -> dict[str, list[list[int]]]:
        """Featurises the input molecules as unfolded MinHash fingerprints.

        Args:
            samples: The array of samples to featurise.
            n_permutations: Analogous to the number of bits ECFP fingerprints
                are folded into. Higher is better, lower is less exact. Defaults
                to 1024.
            radius: Analogous to the radius for the Morgan fingerprint. The
                default radius 3 encodes SMILES to MHFP6. Defaults to 3.
            rings: Whether rings in the molecule are included in the fingerprint.
                As a radii of 3 fails to encode rings and there is no way to
                determine ring-membership in a substructure SMILES, this
                considerably increases performance. Defaults to `True`.
            kekulise: Whether or not to kekulise the molecule before extracting
                substructure SMILES. Defaults to `True`.
            sanitise: Whether or not to sanitise the molecule (using RDKit)
                before extracting substructure SMILES. Defaults to `True`.
            seed: The seed for the MinHash operation. Has to be the same for
                comparable fingerprints. Defaults to 42.

        Returns:
            MHFP fingerpints.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation("mhfp_unfolded")
            >>> samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]
            >>> representation.featurise(samples, n_permutations=4)
            {'mhfp_unfolded': [[15876248, 33988699, 14316170, 98479663]]}
        """

        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        encoder = MHFPEncoder(n_permutations=n_permutations, seed=seed)

        fingerprints: list[list[int]] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smiles = to_smiles(sample)
                fingerprint = encoder.encode(
                    smiles,
                    radius=radius,
                    rings=rings,
                    kekulize=kekulise,
                    sanitize=sanitise,
                ).tolist()
                fingerprints.append(fingerprint)

        return {self.tag: fingerprints}



"""Pearson correlation coefficient score function."""

import logging
from typing import Any, Literal

import evaluate
import scipy.stats

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Pearson correlation coefficient and p-value for testing non-correlation.

The Pearson correlation coefficient [1] measures the linear relationship between two datasets.
Like other correlation coefficients, this one varies between -1 and +1 with 0 implying no correlation.
Correlations of -1 or +1 imply an exact linear relationship. Positive correlations imply that as
x increases, so does y. Negative correlations imply that as x increases, y decreases.

This function also performs a test of the null hypothesis that the distributions underlying the samples
are uncorrelated and normally distributed. (See Kowalski [3] for a discussion of the effects of non-normality
of the input on the distribution of the correlation coefficient.) The p-value roughly indicates the probability
of an uncorrelated system producing datasets that have a Pearson correlation at least as extreme as the one
computed from these datasets.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    alternative (optional): Defines the alternative hypothesis. Default is 'two-sided'.
        The following options are available:

        * 'two-sided': the correlation is nonzero
        * 'less': the correlation is negative (less than zero)
        * 'greater':  the correlation is positive (greater than zero)

Returns:
    The Pearson correlation coefficient.
    The p-value associated with the chosen alternative.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("pearson")
    >>> predictions = [1, 2, 3, 4, 5]
    >>> references = [5, 6, 7, 8, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'pearson::correlation': 0.83205..., 'pearson::p_value': 0.0805...}
"""

_CITATION = """
“Pearson correlation coefficient”, Wikipedia, https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
"""

Alternative = Literal["two-sided", "less", "greater"]


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class Pearson(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        alternative: Alternative = "two-sided",
        **kwargs: Any,
    ) -> MetricResult:
        correlation, p_value = scipy.stats.pearsonr(
            x=predictions,
            y=references,
            alternative=alternative,
        )
        return {
            f"{self.tag}::correlation": correlation,
            f"{self.tag}::p_value": p_value,
        }



import pytest

from molflux.modelzoo.load import load_from_dict, load_model


def test_raises_not_implemented_for_unknown_model():
    """That a NotImplementedError is raised if attempting to load an unknown
    model."""
    name = "unknown"
    with pytest.raises(NotImplementedError):
        load_model(name=name)


def test_raised_error_for_unknown_model_provides_close_matches():
    """That the error raised when attempting to load an unknown model
    shows possible close matches to the user (if any)."""
    name = "random_forst"
    # This should suggest e.g. ["random_forest_classifier", "random_forest_regressor"]
    with pytest.raises(
        NotImplementedError,
        match="You might be looking for one of these",
    ):
        load_model(name=name)


def test_dict_missing_required_fields_raises():
    """That cannot load a model with a config missing required fields."""
    config = {"unknown_key": "value"}
    with pytest.raises(SyntaxError):
        load_from_dict(config)



from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Literal

import h5py

import datasets
from molflux.datasets.typing import ExamplesGenerator

from .ani2x_configs import FEATURES, LEVEL_OF_THEORY, URL_DICT

_BASE_URL = "https://zenodo.org/records/10108942"

_HOMEPAGE = "https://pubs.acs.org/doi/10.1021/acs.jctc.0c00121"

_DESCRIPTION = """
    The new model, dubbed ANI-2x, is trained to three additional chemical elements: S, F, and Cl.
    Additionally, ANI-2x underwent torsional refinement training to better predict molecular torsion
    profiles. These new features open a wide range of new applications within organic chemistry and
    drug development. These seven elements (H, C, N, O, F, Cl, and S) make up ~90% of drug-like molecules.

    WARNING: The molecules here are point clouds. The OpenEye mols do not have any bonds.
    """


@dataclass
class ANI2XConfig(datasets.BuilderConfig):
    backend: Literal["openeye", "rdkit"] = "rdkit"
    level_of_theory: LEVEL_OF_THEORY = "wB97X/631Gd"


class ANI2X(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = ANI2XConfig
    config: ANI2XConfig

    BUILDER_CONFIGS: ClassVar[list[datasets.BuilderConfig]] = [
        ANI2XConfig(
            name="rdkit",
            backend="rdkit",
            level_of_theory="wB97X/631Gd",
        ),
        ANI2XConfig(
            name="openeye",
            backend="openeye",
            level_of_theory="wB97X/631Gd",
        ),
    ]

    def _info(self) -> datasets.DatasetInfo:
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(FEATURES[self.config.level_of_theory]),
            homepage=_HOMEPAGE,
        )

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        archive_path = dl_manager.download_and_extract(
            URL_DICT[self.config.level_of_theory]["url"],
        )

        archive_path = str(
            Path(archive_path) / URL_DICT[self.config.level_of_theory]["path"],
        )
        return [
            datasets.SplitGenerator(
                name="train",
                gen_kwargs={
                    "h5_file_path": archive_path,
                },
            ),
        ]

    def _generate_examples(self, **kwargs: Any) -> ExamplesGenerator:
        if self.config.backend == "openeye":
            return self._generate_examples_openeye(**kwargs)
        elif self.config.backend == "rdkit":
            return self._generate_examples_rdkit(**kwargs)
        else:
            raise KeyError("Unknown backend")

    def _generate_examples_openeye(
        self,
        h5_file_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from openeye.oechem import OEMol, OEWriteMolToBytes
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("openeye-toolkits", e) from e

        # open h5 file
        h5_file = h5py.File(h5_file_path, "r+")

        # get columns to loop over
        columns = list(self._info().features.keys())
        columns.remove("mol_bytes")

        # reset index
        index = 0
        for _num_atoms, properties in h5_file.items():
            properties_list = [
                dict(zip(properties, t, strict=False))
                for t in zip(*properties.values(), strict=False)
            ]

            for prop in properties_list:
                mol = OEMol()
                for atom_num in prop["species"]:
                    mol.NewAtom(int(atom_num))
                flat_coords = [float(xs) for x in prop["coordinates"] for xs in x]
                mol.SetCoords(flat_coords)

                mol_bytes = OEWriteMolToBytes(".oeb", mol)

                example_dict = {
                    "mol_bytes": mol_bytes,
                }
                for feature in columns:
                    example_dict[feature] = prop[feature]

                index += 1
                yield index - 1, example_dict

    def _generate_examples_rdkit(
        self,
        h5_file_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from rdkit.Chem import Atom, Conformer, RWMol
            from rdkit.Geometry import Point3D
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("rdkit", e) from e

        # open h5 file
        h5_file = h5py.File(h5_file_path, "r+")

        # get columns to loop over
        columns = list(self._info().features.keys())
        columns.remove("mol_bytes")

        # reset index
        index = 0
        for _num_atoms, properties in h5_file.items():
            properties_list = [
                dict(zip(properties, t, strict=False))
                for t in zip(*properties.values(), strict=False)
            ]

            for prop in properties_list:
                mol = RWMol()
                for _idx, atomic_num in enumerate(prop["species"]):
                    atom = Atom(int(atomic_num))
                    mol.AddAtom(atom)

                conf = Conformer()
                for i in range(mol.GetNumAtoms()):
                    conf.SetAtomPosition(i, Point3D(*(prop["coordinates"][i].tolist())))

                mol.AddConformer(conf)

                mol_bytes = mol.ToBinary()

                example_dict = {
                    "mol_bytes": mol_bytes,
                }
                for feature in columns:
                    example_dict[feature] = prop[feature]

                index += 1
                yield index - 1, example_dict



import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "scaffold_rdkit"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_dataset():
    return [
        "c1ccc(cc1)C(C#N)OC2C(C(C(C(O2)COC3C(C(C(C(O3)CO)O)O)O)O)O)O",
        "Cc1c(cco1)C(=O)Nc2ccccc2",
        "CC(=CCCC(=CC=O)C)C",
        "c1ccc2c(c1)ccc3c2ccc4c3ccc5c4cccc5",
        "c1ccsc1",
        "c1ccc2c(c1)ncs2",
        "c1cc(c(c(c1)Cl)c2c(cc(cc2Cl)Cl)Cl)Cl",
        "CC12CCC3c4ccc(cc4CCC3C1CCC2O)O",
        "C1C2C3C(C1C4C2O4)C5(C(=C(C3(C5(Cl)Cl)Cl)Cl)Cl)Cl",
        "CC(=C)C1Cc2c(ccc3c2OC4COc5cc(c(cc5C4C3=O)OC)OC)O1",
        "C1CC(=O)NC1",
        "c1ccc2cc(ccc2c1)Cl",
        "CCCC=C",
        "CCC1(C(=O)NCNC1=O)c2ccccc2",
        "CCCCCCCCCCCCCC",
        "CC(C)Cl",
        "CCC(C)CO",
        "c1ccc(cc1)C#N",
        "CCOP(=S)(OCC)Oc1cc(nc(n1)C(C)C)C",
        "CCCCCCCCCC(C)O",
        "c1cc(c(cc1Cl)Cl)c2c(ccc(c2Cl)Cl)Cl",
        "C1CCC(CC1)n2c(=O)c3c([nH]c2=O)CCC3",
        "CCOP(=S)(OCC)SCSCC",
        "CCOc1ccc(cc1)NC(=O)C",
        "CCN(CC)c1c(cc(c(c1N(=O)=O)N)C(F)(F)F)N(=O)=O",
        "CCCCCCCO",
        "Cn1c2c(c(=O)n(c1=O)C)[nH]cn2",
        "CCCCC1(C(=O)NC(=O)NC1=O)CC",
        "c1cc(ccc1C(=C(Cl)Cl)c2ccc(cc2)Cl)Cl",
        "CCCCCCCC(=O)OC",
        "CCc1ccc(cc1)CC",
        "CCOP(=S)(OCC)SCSC(C)(C)C",
        "Cc1cccc(c1)NC(=O)Oc2cccc(c2)NC(=O)OC",
        "C=C(Cl)Cl",
        "Cc1cccc-2c1Cc3c2cccc3",
        "CCCCC=O",
        "c1ccc(cc1)Nc2ccccc2",
        "CN(C)C(=O)SCCCCOc1ccccc1",
        "CCCOP(=S)(OCCC)SCC(=O)N1CCCCC1C",
        "CCCCCCCI",
        "c1ccc(cc1)c2cccc(c2)Cl",
        "C=CCCCO",
        "C1CC2(C1)C(=O)NC(=O)NC2=O",
        "CC1CCC(C(C1)O)C(C)C",
        "CC(C)OC=O",
        "CCCCCC(C)O",
        "CC(=O)Nc1ccc(cc1)Br",
        "c1ccc(cc1)n2c(=O)c(c(cn2)N)Br",
        "CC1=C(C(C(=C(N1)C)C(=O)OC)c2ccccc2N(=O)=O)C(=O)OC",
        "Cc1ccc2ccc(nc2c1)C",
        "CCCCCCC#C",
        "CCC1(C(=O)NC(=O)NC1=O)C2=CCCCC2",
        "c1ccc2c(c1)ccc3c2ccc4c3cccc4",
        "CCC(C)n1c(=O)c(c([nH]c1=O)C)Br",
        "c1cc(c(c(c1)Cl)Cl)c2c(c(cc(c2Cl)Cl)Cl)Cl",
        "Cc1ccccc1O",
        "CC(C)CCC(C)(C)C",
        "Cc1ccc(c2c1cccc2)C",
        "Cc1cc2c3ccccc3ccc2c4c1cccc4",
        "CCCC(=O)C",
        "c1c(c(c(c(c1Cl)Cl)c2c(c(cc(c2Cl)Cl)Cl)Cl)Cl)Cl",
        "CCCOC(=O)CC",
        "CC12CC(C3(C(C1CC(C2(C(=O)CO)O)O)CCC4=CC(=O)C=CC43C)F)O",
        "c1cc(ccc1N)O",
        "c1ccc(cc1)CNC(=O)Cn2ccnc2N(=O)=O",
        "c1ccc2c(c1)C(=O)C(=C(C2=O)O)C3CCC(CC3)c4ccc(cc4)Cl",
        "CCNc1nc(nc(n1)Cl)N(CC)CC",
        "c1cnc(cn1)C(=O)N",
        "CCC(CC)(C(=O)NC(=O)N)Br",
        "c1ccc(c(c1)c2ccccc2Cl)Cl",
        "c1cc(oc1C=NN2CC(=O)NC2=O)N(=O)=O",
        "c1cc(ccc1N(=O)=O)Oc2ccc(cc2Cl)Cl",
        "CC1(C2CCC1(C(=O)C2)C)C",
        "C=CCC1(C(=O)NC(=O)NC1=O)c2ccccc2",
        "CCCCC(=O)OCC",
        "CC(C)CCOC(=O)C",
        "CCCCCC(=O)OCN1C(=O)C(NC1=O)(c2ccccc2)c3ccccc3",
        "c1cc(cc(c1)Cl)c2cc(ccc2Cl)Cl",
        "CCCBr",
        "CCCC1COC(O1)(Cn2cncn2)c3ccc(cc3Cl)Cl",
        "CN(C=O)C(=O)CSP(=S)(OC)OC",
        "Cc1c2c(nccn2)ncn1",
        "C(=S)(N)N",
        "Cc1ccc(cc1)C",
        "CCc1ccccc1CC",
        "C(C(Cl)(Cl)Cl)(Cl)(Cl)Cl",
        "CC(C)C(c1ccc(cc1)OC(F)F)C(=O)OC(C#N)c2cccc(c2)Oc3ccccc3",
        "CCCN(=O)=O",
        "CC1CCC(C(=O)C1)C(C)C",
        "CCN1c2cc(ccc2NC(=O)c3c1nccc3)Cl",
        "c1cc(c(c(c1)Cl)Cl)N(=O)=O",
        "CCCC(C)C1(C(=O)NC(=S)NC1=O)CC=C",
        "c1ccc-2c(c1)-c3cccc4c3c2ccc4",
        "CCCOC(C)C",
        "Cc1cc(c2ccccc2c1)C",
        "CCC(=C(CC)c1ccc(cc1)O)c2ccc(cc2)O",
        "C(#N)c1c(c(c(c(c1Cl)Cl)Cl)C#N)Cl",
        "c1cc(c(cc1c2cc(ccc2Cl)Cl)Cl)Cl",
        "c1ccc(cc1)C2CO2",
        "CC(C)c1ccccc1",
        "CC12CCC3C(C1CCC2C(=O)CO)CCC4=CC(=O)CCC34C",
        "c1cnc2c(n1)c(c(c(c2Cl)Cl)Cl)Cl",
        "C1C(C(C(C(O1)O)O)O)O",
        "C(Cl)Cl",
        "CCc1cccc2c1cccc2",
        "COC=O",
        "c1ccc(c(c1)N(=O)=O)O",
        "Cc1c[nH]c(=O)[nH]c1=O",
        "CC(C)C",
        "c1nc2c(c(n1)O)ncn2C3C(C(C(O3)CO)O)O",
        "c1c(cc(c(c1I)O)I)C#N",
        "c1cc(c(cc1N(=O)=O)Cl)NC(=O)c2cc(ccc2O)Cl",
        "CCCCC",
        "c1ccc(cc1)O",
        "c1ccc2cc3cc(ccc3cc2c1)N",
        "Cn1cnc2c1c(=O)[nH]c(=O)n2C",
        "c1ccc2cnccc2c1",
        "CC(C)N(c1ccc(cc1)Cl)C(=O)CSP(=S)(OC)OC",
        "CCCCCCc1ccccc1",
        "c1ccc(cc1)c2ccccc2Cl",
    ]


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_yields_one_fold_by_default(fixture_sample_dataset, fixture_test_strategy):
    """That the splitting strategy only yields one set of splits by default."""
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=range(len(dataset)), y=dataset)
    assert len(list(indices)) == 1


def test_default_split_fractions(fixture_sample_dataset, fixture_test_strategy):
    """That the dataset is split by default into 80:10:10 splits."""
    dataset = fixture_sample_dataset
    n_samples = len(dataset)
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=range(len(dataset)), y=dataset)
    train_indices, validation_indices, test_indices = next(indices)
    assert len(list(train_indices)) == 0.8 * n_samples
    assert len(list(validation_indices)) == 0.1 * n_samples
    assert len(list(test_indices)) == 0.1 * n_samples


def test_splits_are_disjoint(fixture_sample_dataset, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=range(len(dataset)), y=dataset)
    train_indices, validation_indices, test_indices = next(indices)
    assert set(train_indices).isdisjoint(set(validation_indices))
    assert set(train_indices).isdisjoint(set(test_indices))
    assert set(validation_indices).isdisjoint(set(test_indices))


def test_inconsistent_split_fractions_raise(
    fixture_sample_dataset,
    fixture_test_strategy,
):
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=range(len(dataset)),
        y=dataset,
        train_fraction=0.6,
        validation_fraction=0.1,
        test_fraction=0.1,
    )
    with pytest.raises(AssertionError):
        next(indices)



from molflux.modelzoo.protocols import Model
from molflux.modelzoo.store.artefact import pack, unpack
from molflux.modelzoo.store.manager import (
    load_artefact_from_store,
    save_artefact_to_store,
)
from molflux.modelzoo.typing import PathLike


def save_to_store(key: PathLike, model: Model) -> None:
    """Saves a pre-trained modelzoo model to a model store."""

    # Pack modelzoo model into artefact
    artefact = pack(model=model)

    # Store the artefact in the store
    save_artefact_to_store(artefact=artefact, directory=str(key))


def load_from_store(key: PathLike) -> Model:
    """Loads a pre-trained modelzoo model from a model store."""

    # Generate a modelzoo artefact from the store
    artefact = load_artefact_from_store(directory=str(key))

    #  get model from artefact
    model = unpack(artefact=artefact)
    return model



import os
from dataclasses import dataclass
from typing import Any, ClassVar, Literal

import pandas as pd

import datasets
from molflux.datasets.typing import ExamplesGenerator

_BASE_URL_DICT = {
    "mols_file": "http://ogb-data.stanford.edu/data/lsc/pcqm4m-v2-train.sdf.tar.gz",
    "dataset_file": "https://dgl-data.s3-accelerate.amazonaws.com/dataset/OGB-LSC/pcqm4m-v2.zip",
}

_HOMEPAGE = "https://ogb.stanford.edu/docs/lsc/pcqm4mv2/"

_DESCRIPTION = """
   PCQM4Mv2 is a quantum chemistry dataset originally curated under the PubChemQC project.
   Based on the PubChemQC, we define a meaningful ML task of predicting DFT-calculated HOMO-LUMO
   energy gap of molecules given their 2D molecular graphs. The HOMO-LUMO gap is one of the most
   practically-relevant quantum chemical properties of molecules since it is related to reactivity,
   photoexcitation, and charge transport. Moreover, predicting the quantum chemical property only from
   2D molecular graphs without their 3D equilibrium structures is also practically favorable. This is
   because obtaining 3D equilibrium structures requires DFT-based geometry optimization, which is
   expensive on its own.

   We provide molecules as the SMILES strings, from which 2D molecule graphs (nodes are atoms and edges
   are chemical bonds). We further provide the equilibrium 3D graph structure for training molecules.
    """


@dataclass
class PCQM4MV2Config(datasets.BuilderConfig):
    backend: Literal["openeye", "rdkit"] = "rdkit"


class PCQM4MV2(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = PCQM4MV2Config
    config: PCQM4MV2Config

    BUILDER_CONFIGS: ClassVar[list[datasets.BuilderConfig]] = [
        PCQM4MV2Config(
            name="openeye",
            backend="openeye",
        ),
        PCQM4MV2Config(
            name="rdkit",
            backend="rdkit",
        ),
    ]

    def _info(self) -> datasets.DatasetInfo:
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "mol_bytes": datasets.Value("binary"),
                    "smiles": datasets.Value("string"),
                    "homo_lumo_gap": datasets.Value("float64"),
                },
            ),
            homepage=_HOMEPAGE,
        )

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        raw_file_dict = dl_manager.download_and_extract(_BASE_URL_DICT)

        mols_file = os.path.join(raw_file_dict["mols_file"], "pcqm4m-v2-train.sdf")
        csv_file = dl_manager.extract(
            os.path.join(raw_file_dict["dataset_file"], "pcqm4m-v2/raw/data.csv.gz"),
        )

        return [
            datasets.SplitGenerator(
                name="train",
                gen_kwargs={
                    "mols_file": mols_file,
                    "csv_file": csv_file,
                },
            ),
        ]

    def _generate_examples(self, **kwargs: Any) -> ExamplesGenerator:
        if self.config.backend == "openeye":
            return self._generate_examples_openeye(**kwargs)
        elif self.config.backend == "rdkit":
            return self._generate_examples_rdkit(**kwargs)
        else:
            raise KeyError("Unknown backend")

    def _generate_examples_openeye(
        self,
        mols_file: str | None = None,
        csv_file: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from openeye.oechem import (
                OEErrorLevel_Error,
                OEThrow,
                OEWriteMolToBytes,
                oemolistream,
            )
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("openeye-toolkits", e) from e

        OEThrow.SetLevel(OEErrorLevel_Error)

        df = pd.read_csv(csv_file)

        # get rid of points with no homolumogap value
        df = df[~df["homolumogap"].isna()][["smiles", "homolumogap"]]
        df = df.rename(columns={"homolumogap": "homo_lumo_gap"})

        # find mol iterator
        mols_iterator = oemolistream(mols_file).GetOEGraphMols()

        for index, (mol, (_, row)) in enumerate(
            zip(mols_iterator, df.iterrows(), strict=False),
        ):
            row_dict = row.to_dict()
            row_dict["mol_bytes"] = OEWriteMolToBytes(".oeb", mol)

            yield index, row_dict

    def _generate_examples_rdkit(
        self,
        mols_file: str | None = None,
        csv_file: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from rdkit import Chem
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("rdkit", e) from e

        df = pd.read_csv(csv_file)

        # get rid of points with no homolumogap value
        df = df[~df["homolumogap"].isna()][["smiles", "homolumogap"]]
        df = df.rename(columns={"homolumogap": "homo_lumo_gap"})

        # find mol iterator
        supplier = Chem.SDMolSupplier(
            mols_file,
            sanitize=False,
            removeHs=False,
        )

        for index, (mol, (_, row)) in enumerate(
            zip(supplier, df.iterrows(), strict=False),
        ):
            row_dict = row.to_dict()
            row_dict["mol_bytes"] = mol.ToBinary()

            yield index, row_dict



import pytest

from molflux.metrics.load import load_suite
from molflux.metrics.metric import Metrics
from molflux.metrics.suites.catalogue import list_suites

suite_name = "uncertainty"


def test_in_catalogue():
    """That the suite is in the catalogue."""
    suites = list_suites()
    assert suite_name in suites


def test_loads():
    """That the suite can be loaded without errors."""
    assert load_suite(name=suite_name)


def test_is_metrics():
    """That the suite returns a metrics collection."""
    suite = load_suite(name=suite_name)
    assert isinstance(suite, Metrics)


def test_results_have_all_the_metrics_in():
    """Check that results have the uncertainty metrics."""
    ground_truth = [0, 0.3, 0.5, 0.8, 1]
    preds = [0.1, 0.35, 0.45, 0.68, 1.2]
    intervals = [(0, 0.5), (0.1, 0.4), (0.2, 0.5), (0.5, 0.7), (0.7, 1.5)]
    uq_suite = load_suite(name=suite_name)
    result = uq_suite.compute(
        references=ground_truth,
        predictions=preds,
        prediction_intervals=intervals,
    )
    all_metrics = [metric.name for metric in uq_suite]
    assert set(result.keys()) - set(all_metrics) == set()
    assert set(all_metrics) - set(result.keys()) == set()


def test_compute_works():
    """Check that computations work for a few of the
    metrics in the uncertainty suite."""
    ground_truth = [0, 0.3, 0.5, 0.8, 1]
    preds = [0.1, 0.35, 0.45, 0.68, 1.2]
    intervals = [(0, 0.5), (0.1, 0.4), (0.2, 0.5), (0.5, 0.7), (0.7, 1.5)]
    uq_suite = load_suite(name=suite_name)
    result = uq_suite.compute(
        references=ground_truth,
        predictions=preds,
        prediction_intervals=intervals,
    )
    assert pytest.approx(result["prediction_interval_width"], 0.01) == 0.42
    assert pytest.approx(result["prediction_interval_coverage"], 0.01) == 0.8



from __future__ import annotations

import pytest
import rdkit

from molflux.features import Representation, list_representations, load_representation
from molflux.features.errors import FeaturisationError
from molflux.features.representations.rdkit.descriptors.rdkit_descriptors_2d import (
    RdkitDescriptors_2d,
    list_available_rdkit_descriptors_2d,
)

representation_name = "rdkit_descriptors_2d"
_EXPECTED_NUM_2D_DESCRIPTORS = 210


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, RdkitDescriptors_2d)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_representation_tag_matches_entrypoint_name(fixture_representation):
    """That the default representation tag matches the entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    representation = fixture_representation
    assert representation.tag == representation_name


def test_number_output_features(fixture_representation):
    """That this one-to-many representations generates the expected number
    of features."""
    representation = fixture_representation
    samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]
    result = representation.featurise(samples)
    assert len(result) == _EXPECTED_NUM_2D_DESCRIPTORS


def test_list_available_rdkit_descriptors_2d_number_output_features():
    """That the function lists out the expected number of available descriptors."""
    available = list_available_rdkit_descriptors_2d()
    assert len(available) == _EXPECTED_NUM_2D_DESCRIPTORS


def test_list_available_rdkit_descriptors_2d_matches_rdkit_backend():
    """That the function matches the 2D descriptors available in rdkit"""
    available = list_available_rdkit_descriptors_2d()
    source = [descriptor[0] for descriptor in rdkit.Chem.Descriptors._descList]
    # remove duplicates to patch a duplicate SPS descriptor in rdkit.Chem.Descriptors._descList (bug)
    source = list(dict.fromkeys(source))
    assert available == source


def test_include_selects_correct_number_of_computed_features(fixture_representation):
    """That the include kwarg can be used to select a specific number of
    descriptors to be calculated."""
    representation = fixture_representation
    samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]
    include = ["AvgIpc", "qed", "Chi0"]
    result = representation.featurise(samples, include=include)
    assert len(result) == len(include)


def test_request_to_include_nonexistent_descriptor_raises(fixture_representation):
    """That an error is raised if requesting an non-existent descriptor."""
    representation = fixture_representation
    samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]

    wrong = "WRONG"
    include = ["AvgIpc", wrong, "Chi0"]

    with pytest.raises(
        ValueError,
        match=r".*The following descriptor\(s\) are not available.*",
    ):
        representation.featurise(samples, include=include)


def test_request_to_include_nonexistent_similar_descriptor_raises_with_helpful_message(
    fixture_representation,
):
    """That an informative error message is returned if requesting slightly incorrect descriptors."""
    representation = fixture_representation
    samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]

    typo = "qeds"  # instead of qed
    include = ["AvgIpc", typo, "Chi0"]
    with pytest.raises(ValueError, match=r".*You might be looking for one of these.*"):
        representation.featurise(samples, include=include)


def test_exclude_deselects_correct_number_of_computed_features(fixture_representation):
    """That the exclude kwarg can be used to deselect a specific number of
    descriptors to be calculated."""
    representation = fixture_representation
    samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]
    exclude = ["AvgIpc", "qed", "Chi0"]
    result = representation.featurise(samples, exclude=exclude)
    assert len(result) == _EXPECTED_NUM_2D_DESCRIPTORS - len(exclude)


def test_no_descriptors_to_calculate_raises(fixture_representation):
    """That an error is raised if no descriptors are left to be calculated.

    This could be because of explicitly not selecting any descriptors, or by
    providing 'select' and 'exclude' arrays that complete filtering out of
    any descriptor.
    """
    representation = fixture_representation
    samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]

    include = ["AvgIpc", "qed", "Chi0"]
    exclude = include

    with pytest.raises(ValueError, match=r".*No descriptors to calculate.*"):
        representation.featurise(samples, include=include, exclude=exclude)


def test_default_compute(fixture_representation):
    """That batch scoring gives expected results."""
    representation = fixture_representation
    samples = ["COc1cc2c(cc1OCCCN3CCOCC3)c(ncn2)Sc4nccs4", "c1ccccc1"]

    include = ["MaxEStateIndex", "qed", "MolWt"]
    result = representation.featurise(samples, include=include)

    expected = {
        "rdkit_descriptors_2d::MaxEStateIndex": [6.0566621787603925, 2.0],
        "rdkit_descriptors_2d::qed": [0.407489782854996, 0.4426283718993647],
        "rdkit_descriptors_2d::MolWt": [418.544, 78.11399999999999],
    }

    # we reserve the right to change key naming conventions in the future
    assert list(result.values()) == list(expected.values())


def test_default_compute_one(fixture_representation):
    """That scoring on single input gives expected results."""
    representation = fixture_representation
    samples = ["COc1cc2c(cc1OCCCN3CCOCC3)c(ncn2)Sc4nccs4"]

    include = ["MaxEStateIndex", "qed", "MolWt"]
    result = representation.featurise(samples, include=include)

    expected = {
        "rdkit_descriptors_2d::MaxEStateIndex": [6.0566621787603925],
        "rdkit_descriptors_2d::qed": [0.407489782854996],
        "rdkit_descriptors_2d::MolWt": [418.544],
    }

    # we reserve the right to change key naming conventions in the future
    assert list(result.values()) == list(expected.values())


def test_default_compute_zero(fixture_representation):
    """That scoring on empty inputs is possible and results in a no-op."""
    representation = fixture_representation
    samples: list[float] = []

    include = ["MaxEStateIndex", "qed", "MolWt"]
    result = representation.featurise(samples, include=include)

    expected: dict[str, list[float]] = {
        "rdkit_descriptors_2d::MaxEStateIndex": [],
        "rdkit_descriptors_2d::qed": [],
        "rdkit_descriptors_2d::MolWt": [],
    }

    # we reserve the right to change key naming conventions in the future
    assert list(result.values()) == list(expected.values())


def test_ordered_results(fixture_representation):
    """That output features align with the order of requested descriptors."""
    representation = fixture_representation
    samples = ["COc1cc2c(cc1OCCCN3CCOCC3)c(ncn2)Sc4nccs4", "c1ccccc1"]

    include_a = ["MaxEStateIndex", "qed", "MolWt"]
    result_a = representation.featurise(samples, include=include_a)

    include_b = include_a[::-1]
    result_b = representation.featurise(samples, include=include_b)

    assert list(result_a.values()) == list(result_b.values())[::-1]
    assert list(result_a.keys()) == list(result_b.keys())[::-1]


@pytest.mark.parametrize(
    ["samples", "index_problematic"],
    [(["CC", "Cc1c(c2ncc(c(n2n1)C(=O)NC3CC(C3)(F)F)N$C)C(=O)N"], 1)],  # MLOPS-991
)
def test_raises_featurisation_error_on_invalid_input_samples(
    fixture_representation,
    samples,
    index_problematic,
):
    """That a FeaturisationError is raised when trying to featurise 'invalid' input samples.

    This ensures that a standardised exception is raised by all representations on internal runtime errors.
    """
    representation = fixture_representation
    with pytest.raises(FeaturisationError) as excinfo:
        representation.featurise(samples)

        # check that a reference to the problematic sample is present in the error message
        assert samples[index_problematic] in str(excinfo.value)



import os
from dataclasses import dataclass
from typing import Any, ClassVar, Literal

import pandas as pd

import datasets
from molflux.datasets.typing import ExamplesGenerator

_BASE_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb9.tar.gz"

_HOMEPAGE = "https://moleculenet.org/"

_DESCRIPTION = """
    QM9 dataset of quantum properties of small molecules with up to 29 atoms and up to 9 heavy atoms.
    """


@dataclass
class GDB9Config(datasets.BuilderConfig):
    backend: Literal["openeye", "rdkit"] = "rdkit"


class GDB9(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = GDB9Config
    config: GDB9Config

    BUILDER_CONFIGS: ClassVar[list[datasets.BuilderConfig]] = [
        GDB9Config(
            name="openeye",
            backend="openeye",
        ),
        GDB9Config(
            name="rdkit",
            backend="rdkit",
        ),
    ]

    def _info(self) -> datasets.DatasetInfo:
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "mol_bytes": datasets.Value("binary"),
                    "mol_id": datasets.Value("string"),
                    "A": datasets.Value("float"),
                    "B": datasets.Value("float"),
                    "C": datasets.Value("float"),
                    "mu": datasets.Value("float"),
                    "alpha": datasets.Value("float"),
                    "homo": datasets.Value("float"),
                    "lumo": datasets.Value("float"),
                    "gap": datasets.Value("float"),
                    "r2": datasets.Value("float"),
                    "zpve": datasets.Value("float"),
                    "u0": datasets.Value("float"),
                    "u298": datasets.Value("float"),
                    "h298": datasets.Value("float"),
                    "g298": datasets.Value("float"),
                    "cv": datasets.Value("float"),
                    "u0_atom": datasets.Value("float"),
                    "u298_atom": datasets.Value("float"),
                    "h298_atom": datasets.Value("float"),
                    "g298_atom": datasets.Value("float"),
                },
            ),
            homepage=_HOMEPAGE,
        )

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        archive_path = dl_manager.download_and_extract(_BASE_URL)

        mols_file = os.path.join(archive_path, "gdb9.sdf")
        properties_file = os.path.join(archive_path, "gdb9.sdf.csv")

        return [
            datasets.SplitGenerator(
                name="train",
                gen_kwargs={
                    "mols_filepath": mols_file,
                    "properties_filepath": properties_file,
                },
            ),
        ]

    def _generate_examples(self, **kwargs: Any) -> ExamplesGenerator:
        if self.config.backend == "openeye":
            return self._generate_examples_openeye(**kwargs)
        elif self.config.backend == "rdkit":
            return self._generate_examples_rdkit(**kwargs)
        else:
            raise KeyError("Unknown backend")

    def _generate_examples_openeye(
        self,
        mols_filepath: str | None = None,
        properties_filepath: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from openeye.oechem import OEIsReadable, OEWriteMolToBytes, oemolistream
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("openeye-toolkits", e) from e

        if not OEIsReadable(mols_filepath):
            raise ValueError(f"{mols_filepath} is not a valid file.")

        mols_iterator = oemolistream(mols_filepath).GetOEGraphMols()

        df = pd.read_csv(properties_filepath)

        for mol, (index, labels) in zip(mols_iterator, df.iterrows(), strict=False):
            example_dict = {
                "mol_bytes": OEWriteMolToBytes(".oeb", mol),
                **{k.lstrip(" "): v for k, v in labels.to_dict().items()},
            }
            yield index, example_dict

    def _generate_examples_rdkit(
        self,
        mols_filepath: str | None = None,
        properties_filepath: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from rdkit import Chem
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("rdkit", e) from e

        supplier = Chem.SDMolSupplier(
            mols_filepath,
            sanitize=False,
            removeHs=False,
        )

        df = pd.read_csv(properties_filepath)

        for mol, (index, labels) in zip(supplier, df.iterrows(), strict=False):
            example_dict = {
                "mol_bytes": mol.ToBinary(),
                **{k.lstrip(" "): v for k, v in labels.to_dict().items()},
            }
            yield index, example_dict



from collections.abc import Iterator
from typing import Any

import numpy as np

from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.strategies.core.shuffle_split import ShuffleSplit
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
Shuffle-Group(s)-Out cross-validation iterator.

Provides randomized train/validation/test indices to split data according to a
third-party provided group. This group information can be used to encode
arbitrary domain specific stratifications of the samples as integers.

For instance the groups could be the year of collection of the samples
and thus allow for cross-validation against time-based splits.

Note: The parameters ``train_fraction``, ``validation_fraction`` and
``test_fraction`` refer to groups, and not to samples, as in ShuffleSplit.
"""


class GroupShuffleSplit(ShuffleSplit):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        n_splits: int = 1,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        seed: int | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y (optional): The target variable for supervised learning problems.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Defaults to 1.
            train_fraction (optional): The proportion of groups to include in the train split.
                Defaults to 0.8.
            validation_fraction: The proportion of groups to include in the validation split.
                Defaults to 0.1.
            test_fraction: The proportion of groups to include in the test split.
                Defaults to 0.1.
            seed (optional): Controls the shuffling applied to the data before applying
                the split. Pass an int for reproducible output across multiple function calls.
                Defaults to None.
        """

        if groups is None:
            raise ValueError("The 'groups' parameter should not be None.")

        classes, group_indices = np.unique(np.array(groups), return_inverse=True)
        for group_train, group_validation, group_test in super()._split(
            dataset=classes,
            y=None,
            groups=None,
            n_splits=n_splits,
            train_fraction=train_fraction,
            validation_fraction=validation_fraction,
            test_fraction=test_fraction,
            seed=seed,
            **kwargs,
        ):
            # these are the indices of classes in the partition
            # invert them into data indices
            train_indices = np.flatnonzero(
                np.in1d(group_indices, group_train),  # type: ignore[arg-type]
            )
            validation_indices = np.flatnonzero(
                np.in1d(group_indices, group_validation),  # type: ignore[arg-type]
            )
            test_indices = np.flatnonzero(
                np.in1d(group_indices, group_test),  # type: ignore[arg-type]
            )

            yield train_indices, validation_indices, test_indices



from collections.abc import Iterator
from typing import Any

import numpy as np
from sklearn.utils import check_random_state

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable
from molflux.splits.utils import partition

_DESCRIPTION = """
Stratified ShuffleSplit.

Provides train/validation/test indices to split data in train/validation/test sets.
This cross-validation object is a merge of StratifiedKFold and
ShuffleSplit, which returns stratified randomized folds. The folds
are made by preserving the percentage of samples for each class.

Note: like the ShuffleSplit strategy, stratified random splits
do not guarantee that all folds will be different, although this is
still very likely for sizeable datasets.

References:
    https://github.com/scikit-learn/scikit-learn/blob/baf828ca126bcb2c0ad813226963621cafe38adb/sklearn/model_selection/_split.py#L1849
"""


class StratifiedShuffleSplit(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        n_splits: int = 1,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        seed: int | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y: The target variable for supervised learning problems.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Defaults to 1.
            train_fraction (optional): The proportion of the dataset to include in the train split.
                Defaults to 0.8.
            validation_fraction: The proportion of the dataset to include in the validation split.
                Defaults to 0.1.
            test_fraction: The proportion of the dataset to include in the test split.
                Defaults to 0.1.
            seed (optional): Controls the shuffling applied to the data before applying
                the split. Pass an int for reproducible output across multiple function calls.
                Defaults to None.

        References:
            An extension of scikit-learn's StratifiedShuffleSplit source code
            https://github.com/scikit-learn/scikit-learn/blob/364c77e04/sklearn/model_selection/_split.py#L1977
        """
        if y is None:
            raise ValueError("The 'y' parameter should not be None.")

        np.testing.assert_almost_equal(
            train_fraction + validation_fraction + test_fraction,
            1.0,
        )

        train_cutoff, validation_cutoff = partition(
            dataset,
            train_fraction,
            validation_fraction,
        )
        n_train, n_validation, n_test = (
            train_cutoff,
            validation_cutoff - train_cutoff,
            len(dataset) - validation_cutoff,
        )

        y = np.asarray(y)
        if y.ndim == 2:
            # for multi-label y, map each distinct row to a string repr
            # using join because str(row) uses an ellipsis if len(row) > 1000
            y = np.array([" ".join(row.astype("str")) for row in y])

        classes, y_indices = np.unique(y, return_inverse=True)
        n_classes = classes.shape[0]

        class_counts = np.bincount(y_indices)
        if np.min(class_counts) < 2:
            raise ValueError(
                "The least populated class in y has only 1"
                " member, which is too few. The minimum"
                " number of groups for any class cannot"
                " be less than 2.",
            )

        if n_train > 0 and n_train < n_classes:
            raise ValueError(
                "The train_size = %d should be greater or "
                "equal to the number of classes = %d" % (n_train, n_classes),
            )

        if n_validation > 0 and n_validation < n_classes:
            raise ValueError(
                "The validation_size = %d should be greater or "
                "equal to the number of classes = %d" % (n_validation, n_classes),
            )

        if n_test > 0 and n_test < n_classes:
            raise ValueError(
                "The test_size = %d should be greater or "
                "equal to the number of classes = %d" % (n_test, n_classes),
            )

        # Find the sorted list of instances for each class:
        # (np.unique above performs a sort, so code is O(n logn) already)
        class_indices = np.split(
            np.argsort(y_indices, kind="mergesort"),
            np.cumsum(class_counts)[:-1],
        )

        rng = check_random_state(seed)
        for _ in range(n_splits):
            # calculate how many of each class should be in each split:
            # we run it on each split so that if there are ties in the
            # class-counts, we break them anew in each iteration
            n_train_per_class = _approximate_mode(class_counts, n_train, rng)
            class_counts_remaining = class_counts - n_train_per_class
            n_validation_per_class = _approximate_mode(
                class_counts_remaining,
                n_validation,
                rng,
            )

            train: list[int] = []
            validation: list[int] = []
            test: list[int] = []

            for i in range(n_classes):
                permutation = rng.permutation(class_counts[i])
                perm_indices_class_i = class_indices[i].take(permutation, mode="clip")

                class_n_train = n_train_per_class[i]
                class_n_validation = n_validation_per_class[i]

                cutoff_train = class_n_train
                cutoff_validation = class_n_train + class_n_validation

                class_train, class_validation, class_test = np.split(
                    perm_indices_class_i,
                    (cutoff_train, cutoff_validation),
                )

                train.extend(class_train)
                validation.extend(class_validation)
                test.extend(class_test)

            train = rng.permutation(train).tolist()
            validation = rng.permutation(validation).tolist()
            test = rng.permutation(test).tolist()

            yield train, validation, test


def _approximate_mode(
    class_counts: np.ndarray,
    n_draws: int,
    rng: np.random.RandomState,
) -> np.ndarray:
    """Computes approximate mode of multivariate hypergeometric.
    This is an approximation to the mode of the multivariate
    hypergeometric given by class_counts and n_draws.
    It shouldn't be off by more than one.
    It is the mostly likely outcome of drawing n_draws many
    samples from the population given by class_counts.

    References:
        https://github.com/scikit-learn/scikit-learn/blob/baf828ca126bcb2c0ad813226963621cafe38adb/sklearn/utils/__init__.py#L1021
    """

    # this computes a bad approximation to the mode of the
    # multivariate hypergeometric given by class_counts and n_draws
    continuous = class_counts / class_counts.sum() * n_draws
    # floored means we don't overshoot n_samples, but probably undershoot
    floored: np.ndarray = np.floor(continuous)
    # we add samples according to how much "left over" probability
    # they had, until we arrive at n_samples
    need_to_add = int(n_draws - floored.sum())
    if need_to_add > 0:
        remainder = continuous - floored
        values = np.sort(np.unique(remainder))[::-1]
        # add according to remainder, but break ties
        # randomly to avoid biases
        for value in values:
            (inds,) = np.where(remainder == value)
            # if we need_to_add less than what's in inds
            # we draw randomly from them.
            # if we need to add more, we add them all and
            # go to the next value
            add_now = min(len(inds), need_to_add)
            inds = rng.choice(inds, size=add_now, replace=False)
            floored[inds] += 1
            need_to_add -= add_now
            if need_to_add == 0:
                break

    return floored.astype(int)



"""Proportion of estimates within a factor of n of the true value"""

import logging
from typing import Any

import evaluate
import numpy as np

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Assessment of the proportion of estimates that lie within a factor of n (n-fold) of the true value
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    n_fold (optional): Float defining the factor within which predictions are assessed
        compared to references. Defaults to 3.0
    log_scaled_inputs (optional): Bool indicating whether inputs are already log-transformed. If false will
        perform log transform, but assumes inputs in the positive reals. Defaults to False

Returns:
    proportion_within_fold: proportion of predictions that lie within n-fold region

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("proportion_within_fold")
    >>> predictions = [2.5, 4.0, 2, 8]
    >>> references = [3, 0.5, 2, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'proportion_within_fold': 0.75}
"""

_CITATION = """\
@article{ring2011phrma,
  title={PhRMA CPCDC initiative on predictive models of human pharmacokinetics, part 3: comparative assessement of prediction methods of human clearance},
  author={Ring, Barbara J and Chien, Jenny Y and Adkison, Kimberly K and Jones, Hannah M and Rowland, Malcolm and Jones, Rhys Do and Yates, James WT and Ku, M Sherry and Gibson, Christopher R and He, Handan and others},
  journal={Journal of pharmaceutical sciences},
  volume={100},
  number={10},
  pages={4090--4110},
  year={2011},
  publisher={Wiley Online Library}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class ProportionWithinFold(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float64"),
                    "references": datasets.Value("float64"),
                },
            ),
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        n_fold: float = 3.0,
        log_scaled_inputs: bool = False,
        **kwargs: Any,
    ) -> MetricResult:
        pred = np.array(predictions)
        ref = np.array(references)
        if not log_scaled_inputs:
            if np.any(pred <= 0) or np.any(ref <= 0):
                raise RuntimeError(
                    "Proportion within fold used with negative predictions or references, but may not make sense",
                )
            pred = np.log10(pred)
            ref = np.log10(ref)
        abs_errors = np.abs(pred - ref)
        score = sum(abs_errors <= np.log10(n_fold)) / ref.shape[0]
        return {self.tag: score}



import pytest

import molflux.splits.config
from molflux.splits.info import SplittingStrategyInfo


def test_description_required():
    """That a 'description' field is required on initialisation."""
    required = "description"
    with pytest.raises(
        TypeError,
        match=f"required positional argument: * {required!r}",
    ):
        SplittingStrategyInfo()  # type: ignore[call-arg]


def test_to_dict():
    """That can dump the SplittingStrategyInfo object as dict."""
    info = SplittingStrategyInfo(description="test")
    info_dict = info.to_dict()
    assert "description" in info_dict
    assert info_dict.get("description") == "test"


def test_write_to_directory_creates_expected_file(tmp_path):
    """That can dump the SplittingStrategyInfo object as file."""
    info = SplittingStrategyInfo(description="test")
    directory = tmp_path / "info"

    directory.mkdir()
    assert len(list(directory.glob("**/*"))) == 0

    info.write_to_directory(directory=str(directory))
    assert len(list(directory.glob("**/*"))) == 1

    expected_info_filename = (
        directory / molflux.splits.config.SPLITTING_STRATEGY_INFO_FILENAME
    )
    assert expected_info_filename.is_file()



from typing import Any

try:
    from rdkit.Chem import rdFingerprintGenerator

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The atom-pair fingerprint for a molecule.

The atom-pair (and related topological-torsion) fingerprints, are implemented
based on the original paper. Atoms are typed based on atomic number,
number of pi electrons, and the degree of the atom. Optionally information
about atomic chirality can also be integrated into the atom types. Both
fingerprint types can be generated in explicit or sparse form and as bit or
count vectors. These fingerprint types are different from the others in the
RDKit in that bits in the sparse form of the fingerprint can be directly
explained (i.e. the “hashing function” used is fully reversible).

These fingerprints were originally “intended” to be used in count-vectors and
they seem to work better that way. The default behavior of the explicit
bit-vector forms of both fingerprints is to use a “count simulation” procedure
where multiple bits are set for a given feature if it occurs more than once.
The default behavior is to use 4 fingerprint bits for each feature
(so a 2048 bit fingerprint actually stores information about the same number of
features as a 512 bit fingerprint that isn`t using count simulation). The bins
correspond to counts of 1, 2, 4, and 8. As an example of how this works: if a
feature occurs 5 times in a molecule, the bits corresponding to counts 1, 2,
and 4 will be set.

The algorithm used is described here: R.E. Carhart, D.H. Smith, R. Venkataraghavan;
"Atom Pairs as Molecular Features in Structure-Activity Studies: Definition and Applications"
JCICS 25, 64-73 (1985).
"""


class AtomPair(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        min_distance: int = 1,
        max_distance: int = 30,
        include_chirality: bool = False,
        use_2d: bool = True,
        count_simulation: bool = True,
        count_bounds: object | None = None,
        fp_size: int = 2048,
        atom_invariants_generator: object | None = None,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """
        Generates a folded atom-pair fingerprint for each input molecule from a fingerprint generator.

        Args:
            samples: The molecules to be fingerprinted.
            min_distance: minimum distance between atoms to be considered in a pair.
                Default is 1 bond.
            max_distance: maximum distance between atoms to be considered in a pair.
                Default is maxPathLen-1 bonds.
            include_chirality: if set, chirality will be used in the atom
                invariants, this is ignored if atom_invariants_generator is provided.
                Defaults to `False`.
            use_2d: if set, the 2D (topological) distance matrix  will be used.
                Defaults to `True`.
            count_simulation:  if set, use count simulation while  generating
                the fingerprint. Defaults to `True`.
            count_bounds: boundaries for count simulation, corresponding bit
                will be  set if the count is higher than the number provided
                for that spot. Defaults to `None`.
            fp_size: size of the generated fingerprint, does not affect the
                sparse versions. Defaults is 2048.
            atom_invariants_generator: atom invariants to be used during
                fingerprint generation. Defaults to `None`.

        Returns:
            Folded atom-pair fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('atom_pair')
            >>> samples = ['CN1C=NC2=C1C(=O)N(C(=O)N2C)C']
            >>> representation.featurise(samples, fp_size=16)
            {'atom_pair': [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        apgen = rdFingerprintGenerator.GetAtomPairGenerator(
            minDistance=min_distance,
            maxDistance=max_distance,
            includeChirality=include_chirality,
            use2D=use_2d,
            countSimulation=count_simulation,
            countBounds=count_bounds,
            fpSize=fp_size,
            atomInvariantsGenerator=atom_invariants_generator,
        )

        atom_pairs_fp_list: list[Fingerprint] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smiles = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smiles)
                fingerprint = list(apgen.GetFingerprint(mol))
                atom_pairs_fp_list.append(fingerprint)

        return {self.tag: atom_pairs_fp_list}



import difflib
from pathlib import Path

SUITES_CATALOGUE: dict[str, Path] = {}


def get_suite_path(suite_name: str) -> Path:
    """Returns the path to the given suite from the catalogue."""
    suite_path = SUITES_CATALOGUE.get(suite_name)

    if suite_path is None:
        msg = f"Suite {suite_name!r} is not available."
        similar = difflib.get_close_matches(suite_name, SUITES_CATALOGUE.keys())
        if similar:
            msg += f" You might be looking for one of these: {similar}"
        raise NotImplementedError(msg)

    if not suite_path.is_file():
        raise FileNotFoundError(f"Could not find suite definition file: {suite_path!r}")

    return suite_path


def list_suites() -> list[str]:
    """List all available suites."""
    suite_names = sorted(SUITES_CATALOGUE.keys())
    return suite_names


def register_suite(suite_path: Path) -> None:
    """Registers a suite in the catalogue."""
    file_type = suite_path.suffix
    if file_type not in {".yml", ".yaml"}:
        raise ValueError(f"Invalid suite file format: {file_type!r} (expected .yml)")

    suite_name = suite_path.stem
    if suite_name not in SUITES_CATALOGUE:
        SUITES_CATALOGUE[suite_name] = suite_path



from __future__ import annotations

import numpy as np
import pytest

from molflux.modelzoo.models.sklearn.sklearn_discrete_nb.bayes import (  # type: ignore
    CorrectedNB,
    CoverageNB,
    PipelinePilotNB,
)


@pytest.fixture
def fixture_training_data() -> tuple[np.ndarray, np.ndarray]:
    X = np.random.choice([0, 1], size=(10, 100))
    y = np.random.choice([0, 1], size=10)
    return X, y


def test_corrected_nb_fitting_features_work(fixture_training_data):
    X, y = fixture_training_data

    clf = CorrectedNB()
    clf.fit(X, y)

    n_classes = len(np.unique(y))
    n_features = X.shape[1]
    feat_count = np.zeros(shape=(n_classes, n_features))
    cls_count = np.zeros(shape=(n_classes,))
    for i, c in enumerate(np.unique(y)):
        feat_count[i] = X[y == c].sum(axis=0)
        cls_count[i] = len(y[y == c])

    assert (clf.feature_count_ == feat_count).all()
    assert (clf.class_count_ == cls_count).all()

    # test log prob is correct
    for feat_idx in np.random.randint(0, n_features, size=20):
        Ftot = clf.feature_count_[:, feat_idx].sum()
        pbases = clf.class_count_ / clf.class_count_.sum()
        for cls_idx, Fc in zip(
            range(len(clf.classes_)),
            clf.feature_count_[:, feat_idx],
            strict=False,
        ):
            p_smooth = (Fc + 1) / (Ftot + 1 / (pbases[cls_idx]))
            p_norm_smooth = p_smooth / pbases[cls_idx]
            p_log_norm_smooth = np.log(p_norm_smooth)
            assert np.isclose(
                clf.feature_log_prob_[cls_idx, feat_idx],
                p_log_norm_smooth,
            )


def test_coverage_nb_fitting_features_work(fixture_training_data):
    X, y = fixture_training_data

    clf = CoverageNB()
    clf.fit(X, y)

    n_classes = len(np.unique(y))
    n_features = X.shape[1]
    feat_count = np.zeros(shape=(n_classes, n_features))
    cls_count = np.zeros(shape=(n_classes,))
    for i, c in enumerate(np.unique(y)):
        feat_count[i] = X[y == c].sum(axis=0)
        cls_count[i] = len(y[y == c])

    assert (clf.feature_count_ == feat_count).all()
    assert (clf.class_count_ == cls_count).all()

    # test log prob is correct
    for feat_idx in np.random.randint(0, n_features, size=20):
        Ftot = clf.feature_count_[:, feat_idx].sum()
        pbases = clf.class_count_ / clf.class_count_.sum()
        for cls_idx, Fc in zip(
            range(len(clf.classes_)),
            clf.feature_count_[:, feat_idx],
            strict=False,
        ):
            p_smooth = (Fc + 1) / (Ftot + 1 / (pbases[cls_idx]))
            p_norm_smooth = p_smooth / pbases[cls_idx]
            p_log_norm_smooth = np.log(p_norm_smooth)
            assert np.isclose(
                clf.feature_log_prob_[cls_idx, feat_idx],
                p_log_norm_smooth,
            )


def test_pipelinepilot_nb_fitting_features_work(fixture_training_data):
    X, y = fixture_training_data

    clf = PipelinePilotNB()
    clf.fit(X, y)

    n_classes = len(np.unique(y))
    n_features = X.shape[1]
    feat_count = np.zeros(shape=(n_classes, n_features))
    cls_count = np.zeros(shape=(n_classes,))
    for i, c in enumerate(np.unique(y)):
        feat_count[i] = X[y == c].sum(axis=0)
        cls_count[i] = len(y[y == c])

    assert (clf.feature_count_ == feat_count).all()
    assert (clf.class_count_ == cls_count).all()

    # test log prob is correct
    for feat_idx in np.random.randint(0, n_features, size=20):
        Ftot = clf.feature_count_[:, feat_idx].sum()
        # PBases based on total feature counts rather than class counts.
        pbases = clf.feature_count_.sum(axis=1) / clf.feature_count_.sum()
        for cls_idx, Fc in zip(
            range(len(clf.classes_)),
            clf.feature_count_[:, feat_idx],
            strict=False,
        ):
            p_smooth = (Fc + 1) / (Ftot + 1 / (pbases[cls_idx]))
            p_norm_smooth = p_smooth / pbases[cls_idx]
            p_log_norm_smooth = np.log(p_norm_smooth)
            assert np.isclose(
                clf.feature_log_prob_[cls_idx, feat_idx],
                p_log_norm_smooth,
            )



import inspect
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Any

import numpy as np

import datasets
from molflux.modelzoo.typing import (
    Features,
    PredictionResult,
)


def raise_on_unrecognised_parameters(callable: Callable, **kwargs: Any) -> None:
    """Raises if the callable does not explicitly define one or more of the
    given kwargs.

    Args:
        callable: The callable to shield.
        kwargs: The kwargs that would be passed to the callable.

    Raises:
        ValueError: If one or more kwargs are not defined by the callable.
    """
    expected_signature = inspect.signature(callable)
    expected_parameters = expected_signature.parameters
    parameters_received = set(kwargs.keys())
    if not parameters_received.issubset(expected_parameters):
        f = callable.__name__.lstrip("_")
        unknown_kwargs = list(parameters_received.difference(expected_parameters))
        raise ValueError(
            f"Unknown {f} parameter(s): {unknown_kwargs}\n\n"
            f"Expected signature {f}{expected_signature}",
        )


def validate_features(data: datasets.Dataset, features: Features) -> None:
    """Validate features from the input dataset."""

    if not features:
        raise ValueError("Please specify the model's features.")

    for feature in features:
        if feature not in data.column_names:
            raise ValueError(f"Data does not have required feature: {feature!r}")


def validate_prediction_result_num_tasks(
    prediction_result: PredictionResult,
    y_features: Features,
) -> None:
    """Validates that the prediction result contains results for the expected number of features."""

    if not y_features:
        raise ValueError("Please specify the model's y_features.")

    if len(prediction_result) != len(y_features):
        raise RuntimeError(
            f"Prediction results do not match the expected number of tasks: got {len(prediction_result)}, expected {len(y_features)}",
        )


def pick_features(
    dataset: datasets.Dataset,
    features: str | Features,
) -> datasets.Dataset:
    """Selects multiple columns from a datasets.Dataset"""

    features = [features] if isinstance(features, str) else features

    if set(features).issuperset(dataset.column_names):
        return dataset

    return dataset.select_columns(features)


@contextmanager
def _disabled_huggingface_progress_bars() -> Iterator[None]:
    """Temporarily disables huggingface 'datasets' progress bars,"""
    was_enabled = datasets.is_progress_bar_enabled()

    if not was_enabled:
        yield
        return

    try:
        datasets.disable_progress_bar()
        yield
    finally:
        datasets.enable_progress_bar()


def get_concatenated_array(
    dataset: datasets.Dataset,
    features: str | Features,
) -> np.ndarray:
    """
    Concatenates columns of dataset into a feature vector, ordered by the specified
    features.

    Concatenation is done on the column dimension - number of dataset rows remains the
    same in the output.

    This assumes that features indicated are scalars or lists of scalars, compatible with
    numpy arrays.

    Args:
        dataset (datasets.Dataset): The input dataset whose subset of columns is
            concatenated
        features (str, Features): One or multiple column names that will be
            concatenated in order. They must all be in the input dataset.

    Returns:
        np.array: 2D numpy array of concatenated columns.
    """

    if not len(dataset):
        return np.array([]).reshape(-1, 1)

    if isinstance(features, str):
        features = [features]

    if not set(features).issubset(set(dataset.features)):
        raise KeyError(f"{features=} not a subset of {dataset.features.keys()=}")

    # generate numpy arrays for each feature
    arrays = [np.array(dataset[feature]) for feature in features]

    # compute the maximum dimension found for each feature
    max_dim = max(arr.ndim for arr in arrays)

    # if it's not 1 or 2, something went wrong
    if max_dim not in [1, 2]:
        raise RuntimeError("Features must either be scalars or 1-dimensional!")

    return np.column_stack(arrays)


def format_wrapped_model_tag(tag: str, wrapped_tags: list[str]) -> str:
    """Ensure more informative naming of model tags."""
    wrapped_models_tags = ":".join(wrapped_tags)
    return f"{tag}[{wrapped_models_tags}]"



from collections.abc import Iterable, Iterator
from typing import Any

import sklearn.model_selection

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
K-fold iterator variant with non-overlapping groups.

The same group will not appear in two different folds (the number of
distinct groups has to be at least equal to the number of folds).
The folds are approximately balanced in the sense that the number of
distinct groups is approximately the same in each fold.
"""


class GroupKFold(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        n_splits: int = 2,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y (optional): The target variable for supervised learning problems.
            groups: Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Must be greater than one.
                Defaults to 2.
        """
        # Wrap scikit-learn implementation
        cv = sklearn.model_selection.GroupKFold(
            n_splits=n_splits,
        )

        for indices in cv.split(dataset, y=y, groups=groups):
            train_indices, validation_indices = indices
            test_indices: Iterable[int] = []  # no holdout in cross-validation
            yield train_indices, validation_indices, test_indices



import pytest

from molflux.metrics import load_suite


def test_raises_not_implemented_for_unknown_suite():
    """That a NotImplementedError is raised if attempting to load an unknown
    suite."""
    name = "unknown"
    with pytest.raises(NotImplementedError):
        load_suite(name=name)


def test_raised_error_for_unknown_suite_provides_close_matches():
    """That the error raised when attempting to load an unknown suite
    shows possible close matches to the user (if any)."""
    name = "regres"
    # This should suggest e.g. ["regression"]
    with pytest.raises(
        NotImplementedError,
        match="You might be looking for one of these",
    ):
        load_suite(name=name)



import math

import numpy as np
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "leave_p_groups_out"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_inputs():
    dataset = range(10)
    groups = np.array([1, 1, 2, 2, 2, 3, 3, 3, 3, 3])
    return dataset, groups


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_requires_groups(fixture_test_strategy):
    """That you need to provide groups to use this strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset, p=2))


def test_requires_n_groups(fixture_test_strategy):
    """That you need to provide the p argument to use this strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    groups = [1, 2, 2, 3, 3, 3]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset, groups=groups))


def test_leave_out_more_groups_than_existing_raises(
    fixture_sample_inputs,
    fixture_test_strategy,
):
    """That cannot perform cross validation leaving out more groups than there are."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    n_unique_groups = len(np.unique(groups))
    with pytest.raises(ValueError):
        next(strategy.split(dataset=dataset, groups=groups, p=n_unique_groups + 1))


def test_leave_out_all_existing_groups_raises(
    fixture_sample_inputs,
    fixture_test_strategy,
):
    """That cannot perform cross validation leaving out all groups."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    n_unique_groups = len(np.unique(groups))
    with pytest.raises(ValueError):
        next(strategy.split(dataset=dataset, groups=groups, p=n_unique_groups))


def test_leave_out_no_groups_raises(fixture_sample_inputs, fixture_test_strategy):
    """That cannot perform cross validation leaving out zero groups."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset=dataset, groups=groups, p=0))


def test_expected_result_one_leave_one_out(fixture_test_strategy):
    """That the splitting strategy yields the expected result in this case.

    Note, we are testing that the splits are not shuffled and that groups are
    held out sequentially in the validation set - starting with the first one.
    """
    dataset = range(10)
    groups = [1, 1, 2, 2, 2, 3, 3, 3, 3, 3]
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, p=1)

    train_indices, validation_indices, _ = next(indices)
    assert all(
        i == j for i, j in zip(train_indices, [2, 3, 4, 5, 6, 7, 8, 9], strict=False)
    )
    assert all(i == j for i, j in zip(validation_indices, [0, 1], strict=False))

    train_indices, validation_indices, _ = next(indices)
    assert all(
        i == j for i, j in zip(train_indices, [0, 1, 5, 6, 7, 8, 9], strict=False)
    )
    assert all(i == j for i, j in zip(validation_indices, [2, 3, 4], strict=False))

    train_indices, validation_indices, _ = next(indices)
    assert all(i == j for i, j in zip(train_indices, [0, 1, 2, 3, 4], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices, [5, 6, 7, 8, 9], strict=False)
    )


def test_expected_result_one_leave_two_out(fixture_test_strategy):
    """That the splitting strategy yields the expected result in this case."""
    dataset = range(10)
    groups = [1, 1, 2, 2, 2, 3, 3, 3, 3, 3]
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, p=2)

    train_indices, validation_indices, _ = next(indices)
    assert all(i == j for i, j in zip(train_indices, [5, 6, 7, 8, 9], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices, [0, 1, 2, 3, 4], strict=False)
    )

    train_indices, validation_indices, _ = next(indices)
    assert all(i == j for i, j in zip(train_indices, [2, 3, 4], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices, [0, 1, 5, 6, 7, 8, 9], strict=False)
    )

    train_indices, validation_indices, _ = next(indices)
    assert all(i == j for i, j in zip(train_indices, [0, 1], strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices, [2, 3, 4, 5, 6, 7, 8, 9], strict=False)
    )


def test_does_not_produce_test_indices(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy is a cross validation strategy."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = list(strategy.split(dataset=dataset, groups=groups, p=2))
    assert len(indices) > 1


def test_yields_more_than_one_fold(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy yields several folds.

    We test leaving out the maximum amount of groups, which should result in
    the minimum number of folds generated.
    """
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    max_p = len(np.unique(groups)) - 1
    indices = list(strategy.split(dataset=dataset, groups=groups, p=max_p))
    for train_indices, validation_indices, test_indices in indices:
        assert len(train_indices)
        assert len(validation_indices)
        assert not len(test_indices)


def test_yields_correct_number_of_folds(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy yields the correct number of folds.

    This is expected to be the “N choose p”, as we make folds out of all possible
    combinations of p groups left out, out of N available groups.

    References:
        https://en.wikipedia.org/wiki/Binomial_coefficient
    """
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    n_unique_groups = len(np.unique(groups))
    p = 2
    indices = list(strategy.split(dataset=dataset, groups=groups, p=p))
    assert len(indices) == math.comb(n_unique_groups, p)


def test_splits_are_disjoint(fixture_sample_inputs, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups, p=1)
    for train_indices, validation_indices, _ in indices:
        assert set(train_indices).isdisjoint(set(validation_indices))


def test_each_sample_in_a_different_fold(fixture_sample_inputs, fixture_test_strategy):
    """That each sample ends up in a different fold."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, groups=groups, p=1)
    for train_indices, validation_indices, _ in indices:
        groups_in_train = set(groups[train_indices])
        groups_in_validation = set(groups[validation_indices])
        assert set(groups_in_train).isdisjoint(set(groups_in_validation))


def test_deterministic(fixture_sample_inputs, fixture_test_strategy):
    """That repeated splits are deterministic."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of k-fold splits
    indices_a = strategy.split(dataset, groups=groups, p=2)
    train_indices_a1, validation_indices_a1, _ = next(indices_a)
    train_indices_a2, validation_indices_a2, _ = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, groups=groups, p=2)
    train_indices_b1, validation_indices_b1, _ = next(indices_b)
    train_indices_b2, validation_indices_b2, _ = next(indices_b)

    # Check that we get the same results across folds
    assert all(i == j for i, j in zip(train_indices_a1, train_indices_b1, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a1, validation_indices_b1, strict=False)
    )
    assert all(i == j for i, j in zip(train_indices_a2, train_indices_b2, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a2, validation_indices_b2, strict=False)
    )



from dataclasses import asdict
from typing import Any

import numpy as np
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.mcd import MCD
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
Detecting outliers in a Gaussian distributed dataset using
Minimum Covariance Determinant (MCD): robust estimator of covariance.

The Minimum Covariance Determinant covariance estimator is to be applied
on Gaussian-distributed data, but could still be relevant on data
drawn from a unimodal, symmetric distribution. It is not meant to be used
with multi-modal data (the algorithm used to fit a MinCovDet object is
likely to fail in such a case).
One should consider projection pursuit methods to deal with multi-modal
datasets.

First fit a minimum covariance determinant model and then compute the
Mahalanobis distance as the outlier degree of the data

See :cite:`rousseeuw1999fast,hardin2004outlier` for details.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set,
    i.e. the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.

store_precision : bool
    Specify if the estimated precision is stored.

assume_centered : bool
    If True, the support of the robust location and the covariance
    estimates is computed, and a covariance estimate is recomputed from
    it, without centering the data.
    Useful to work with data whose mean is significantly equal to
    zero but is not exactly zero.
    If False, the robust location and covariance are directly computed
    with the FastMCD algorithm without additional treatment.

support_fraction : float, 0 < support_fraction < 1
    The proportion of points to be included in the support of the raw
    MCD estimate. Default is None, which implies that the minimum
    value of support_fraction will be used within the algorithm:
    [n_sample + n_features + 1] / 2

random_state : int, np.random.Generator instance or None, optional (default=None)
    If int, random_state is the seed used by the random number generator;
    If Generator instance, random_state is the random number generator;
    If None, the random number generator is the Generator instance used
    by `np.random`.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class MCDDetectorConfig(PyODModelConfig):
    contamination: float = 0.1
    store_precision: bool = True
    assume_centered: bool = False
    support_fraction: float | None = None
    random_state: int | np.random.Generator | None = None


class MCDDetector(PyODClassificationMixin, PyODModelBase[MCDDetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[MCDDetectorConfig]:
        return MCDDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> MCD:
        config = self.model_config
        return MCD(
            contamination=config.contamination,
            store_precision=config.store_precision,
            assume_centered=config.assume_centered,
            support_fraction=config.support_fraction,
            random_state=config.random_state,
        )
